#!/usr/bin/env python3
from __future__ import print_function
import os
from os import environ, chmod
import json
import hashlib
import pathlib
import logging
from types import GeneratorType
from collections import Counter, defaultdict
import appdirs
import requests

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

# read environment variables # FIXME not the most modular...

__all__ = ['api_token', 'username', 'group', 'group_to_memfile',
           'idFromShareLink', 'shareLinkFromId',
           'AnnoFetcher', 'Memoizer',
           'HypothesisUtils', 'HypothesisHelper', 'Annotation', 'HypAnnoId']

api_token = environ.get('HYP_API_TOKEN', 'TOKEN')  # Hypothesis API token
username = environ.get('HYP_USERNAME', 'USERNAME') # Hypothesis username
group = environ.get('HYP_GROUP', '__world__')
ucd = appdirs.user_cache_dir()

class JEncode(json.JSONEncoder):
     def default(self, obj):
         if isinstance(obj, tuple):
             return list(obj)
         elif isinstance(obj, HypothesisAnnotation):
             return obj._row

         # Let the base class default method raise the TypeError
         return json.JSONEncoder.default(self, obj)


def group_to_memfile(group, post=lambda group_hash:None):
    if group != '__world__':
        m = hashlib.sha256()
        m.update(group.encode())
        group_hash = m.hexdigest()
    else:
        group_hash = group

    memfile = pathlib.Path(ucd, 'hyputils', f'annos-{group_hash}.json')
    post(group_hash)
    return memfile


def makeSimpleLogger(name, level=logging.INFO):
    # TODO use extra ...
    logger = logging.getLogger(name)
    if logger.handlers:  # prevent multiple handlers
        return logger

    logger.setLevel(level)
    ch = logging.StreamHandler()  # FileHander goes to disk
    fmt = ('[%(asctime)s] - %(levelname)8s - '
           '%(name)14s - '
           '%(filename)16s:%(lineno)-4d - '
           '%(message)s')
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = makeSimpleLogger('hyputils')
logd = log.getChild('data')

if 'CI' not in environ:
    log.debug(' '.join((api_token, username, group)))  # sanity check

# simple uri normalization

def norm(iri):
    if '://' in iri:
        _scheme, iri_norm = iri.split('://', 1)
        if '?hypothesisAnnotationId=' in iri:
            iri_norm, junk = iri.split('?hypothesisAnnotationId=', 1)
    else:
        iri_norm = iri  # the creeping madness has co

    return iri_norm

# annotation retrieval and memoization

class NotOkError(Exception):
    def __init__(self, message, request):
        self.status_code = request.status_code
        self.reason = request.reason
        super().__init__(message)


class AnnoFetcher:
    lsu_default = '1900-01-01T00:00:00.000000+00:00'  # don't need, None is ok
    def __init__(self, api_token=api_token, username=username, group=group):
        if api_token == 'TOKEN':
            log.warning('\x1b[31mWARNING:\x1b[0m NO API TOKEN HAS BEEN SET!')
        self.api_token = api_token
        self.username = username
        self.group = group

    def __call__(self):
        return self.get_annos()

    def h(self):
        return HypothesisUtils(username=self.username, token=self.api_token, group=self.group)

    def yield_from_api(self, search_after=None, limit=None, max_results=None, stop_at=None):
        # use stop before if you want to be evil and hit the api in parallel
        log.info(f'fetching after {search_after}')
        # hard code these to simplify assumptions
        order = 'asc'
        sort = 'updated'
        h = self.h()
        params = {'order':order,
                  'sort':sort,
                  'group':h.group}
        if search_after:
            params['search_after'] = search_after
        if max_results is None and self.group == '__world__':
            log.info(f'searching __world__ as {self.username} since max_results was not set')
            params['user'] = self.username
        if limit is not None:
            params['limit'] = limit

        for row in h.search_all(params, max_results=max_results, stop_at=stop_at):
            yield row

    def get_annos_from_api(self, search_after=None, limit=None, max_results=None, stop_at=None):
        # FIXME batch the memoization of these?
        return [HypothesisAnnotation(r) for r in
                self.yield_from_api(search_after=search_after,
                                    limit=limit,
                                    max_results=max_results,
                                    stop_at=stop_at)]


class Memoizer(AnnoFetcher):  # TODO just use a database ...

    class GroupMismatchError(Exception):
        pass

    def __init__(self, memoization_file=None, api_token=api_token, username=username, group=group):
        super().__init__(api_token=api_token, username=username, group=group)
        if memoization_file is None:
            memoization_file = group_to_memfile(group)
        elif not isinstance(memoization_file, pathlib.Path):
            memoization_file = pathlib.Path(memoization_file)

        self.memoization_file = memoization_file

    def check_group(self, annos):
        if annos:
            group = annos[0].group
            if self.group != group:
                raise self.GroupMismatchError(f'Groups do not match! {self.group} {group}')

    def get_annos_from_file(self):
        jblobs = []
        last_sync_updated = None
        if self.memoization_file is not None:
            try:
                with open(self.memoization_file, 'rt') as f:
                    jblobs_lsu = json.load(f)
                    jblobs, last_sync_updated = jblobs_lsu
                    if isinstance(last_sync_updated, HypothesisAnnotation):
                        jblobs = jblobs_lsu
                        last_sync_updated = last_sync_updated['updated']
                if jblobs is None:
                    raise ValueError('wat')
            except json.decoder.JSONDecodeError:
                with open(self.memoization_file, 'rt') as f:
                    data = f.read()
                if not data:
                    log.info('memoization file exists but is empty')
            except FileNotFoundError:
                log.info('memoization file does not exist')

        annos = [HypothesisAnnotation(jb) for jb in jblobs]
        self.check_group(annos)
        return annos, last_sync_updated

    def add_missing_annos(self, annos, last_sync_updated):
        self.check_group(annos)
        search_after = last_sync_updated
        # start from last_sync_updated because we assume that the websocket is unreliable
        new_annos = self.get_annos_from_api(search_after)  # FIXME batch these
        if not new_annos:
            return annos
        merged = annos + new_annos
        merged_unique = sorted(set(merged), key=lambda a: a.updated)
        dupes = [sorted([a for a in merged_unique if a.id == id], key=lambda a: a.updated)
                 for id, count in Counter(anno.id for anno in merged_unique).most_common()
                 if count > 1]

        will_stay = [dupe[-1] for dupe in dupes]
        to_remove = [a for dupe in dupes for a in dupe[:-1]]

        la = len(annos)
        ld = len(dupes)
        lm = len(merged)
        lmu = len(merged_unique)
        [merged_unique.remove(d) for d in to_remove]  # FIXME in the db context these need to be updates
        lmuc = len(merged_unique)

        log.info(f'added {lmuc - la} new annotations')
        log.info(f'updated {ld} annotations')

        if la != lmuc or dupes:
            self.memoize_annos(merged_unique)

        return merged_unique

    def update_annos_from_api(self, annos, helpers=tuple()):
        """ Assumes these are ordered by updated """
        # FIXME why aren't we just getting the group from the annos??
        self.check_group(annos)
        last_sync_updated = annos[-1].updated
        search_after = last_sync_updated
        # start from last_sync_updated because we assume that the websocket is unreliable
        new_annos = self.get_annos_from_api(search_after)
        if new_annos:
            new_ids = set(a.id for a in new_annos)
            for anno in tuple(annos):  # FIXME memory and perf issues?
                if anno.id in new_ids:
                    annos.remove(anno)

            annos.extend(new_annos)
            # TODO deal with updates
            self.memoize_annos(annos)
            for anno in new_annos:
                for Helper in helpers:
                    Helper(anno, annos)

    def memoize_annos(self, annos):
        # FIXME if there are multiple ws listeners we will have race conditions?
        if self.memoization_file is not None:
            log.info(f'annos updated, memoizing new version with, {len(annos)} members')
            do_chmod = False
            if not self.memoization_file.exists():
                do_chmod = True
                if not self.memoization_file.parent.exists():
                    self.memoization_file.parent.mkdir()

            with open(self.memoization_file, 'wt') as f:
                lsu = annos[-1].updated if annos else None
                alsu = annos, lsu
                json.dump(alsu, f, cls=JEncode)

            if do_chmod:
                chmod(self.memoization_file, 0o600)

        else:
            log.info(f'No memoization file, not saving.')

    def get_annos(self):
        annos, last_sync_updated = self.get_annos_from_file()
        if not annos:
            new_annos = self.get_annos_from_api()
            self.memoize_annos(new_annos)
            return new_annos
        else:
            return self.add_missing_annos(annos, last_sync_updated)

    def add_anno(self, anno, annos):
        annos.append(anno)
        self.memoize_annos(annos)

    def del_anno(self, id_, annos, memoize=True):
        # FIXME this is is SUPER slow
        matches = [a for a in annos if a.id == id_]
        if not matches:
            raise ValueError(f'No annotation with id={id_} could be found.')
        else:
            for match in matches:
                annos.remove(match)
            if memoize:
                self.memoize_annos(annos)

    def update_anno(self, anno, annos):
        self.del_anno(anno.id, annos, memoize=False)
        self.add_anno(anno, annos)

    def update_annos_from_api_response(resp, annos):
        # XXX NOTE this will collide with websocket if unmanaged
        # therefore we will need to somehow sync with that to make sure
        # everything remains sane... maybe a shared log that the
        # websocket handler can peak at an confirm or something
        # (that is going to be a weird boundary to navigate)
        if resp.status_code == 200:
            if resp.request.method == 'GET':
                anno = HypothesisAnnotation(resp.json())
                self.update_anno(anno, annos)
                return anno
            elif resp.request.method == 'POST':
                anno = HypothesisAnnotation(resp.json())
                self.add_anno(anno, annos)
                return anno
            elif resp.request.method == 'PATCH':
                anno = HypothesisAnnotation(resp.json())
                self.update_anno(anno, annos)
                return anno
            elif resp.request.method == 'DELETE':
                id = resp.json()['id']
                self.del_anno(id, annos)


#
# url helpers

def idFromShareLink(link):  # XXX warning this will break
    if 'hyp.is' in link:
        id_ = link.split('/')[3]
        return id_

def shareLinkFromId(id_):
    return 'https://hyp.is/' + id_

# API classes

class HypothesisUtils:
    """ services for authenticating, searching, creating annotations """
    def __init__(self, username=None, token=None, group=None, domain=None, limit=None):
        if domain is None:
            self.domain = 'hypothes.is'
        else:
            self.domain = domain
        if username is not None:
            self.username = username
        if token is not None:
            self.token = token
        self.app_url = 'https://%s/app' % self.domain
        self.api_url = 'https://%s/api' % self.domain
        self.query_url_template = 'https://%s/api/search?{query}' % self.domain
        self.search_url_template = 'https://%s/search?q={query}' % self.domain
        self.group = group if group is not None else '__world__'
        self.single_page_limit = 200 if limit is None else limit  # per-page, the api honors limit= up to (currently) 200
        self.permissions = {
                "read": ['group:' + self.group],
                "update": ['acct:' + self.username + '@hypothes.is'],
                "delete": ['acct:' + self.username + '@hypothes.is'],
                "admin":  ['acct:' + self.username + '@hypothes.is']
                }
        self.ssl_retry = 0

    def authenticated_api_query(self, query_url=None):
        try:
            headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8'}
            r = requests.get(query_url, headers=headers)
            obj = r.json()
            if r.ok:
                return obj
            else:
                raise NotOkError(f'response was not ok! {r.reason} {obj}', r)

        except requests.exceptions.SSLError as e:
            if self.ssl_retry < 5:
                self.ssl_retry += 1
                log.error(f'Ssl error at level {self.ssl_retry} retrying....')
                return self.authenticated_api_query(query_url)
            else:
                self.ssl_retry = 0
                log.error(e)
                return {'ERROR':True, 'rows':tuple()}
        except BaseException as e:
            log.error(e)
            #print('Request, status code:', r.status_code)  # this causes more errors...
            return {'ERROR':True, 'rows':tuple()}

    def make_annotation_payload_with_target_using_only_text_quote(self, url, prefix, exact, suffix,
                                                                  text, tags, document, extra):
        """Create JSON payload for API call."""
        if exact is None:
            target = [{'source':url}]
        else:
            target = [{
                "scope": [url],
                "selector":
                [{
                    "type": "TextQuoteSelector",
                    "prefix": prefix,
                    "exact": exact,
                    "suffix": suffix
                },]
            }]
        if text == None:
            text = ''
        if tags == None:
            tags = []
        if document == None:
            document = {}
        payload = {
            "uri": url,
            "user": 'acct:' + self.username + '@hypothes.is',
            "permissions": self.permissions,
            "group": self.group,
            "target": target,
            "tags": tags,
            "text": text,
            "document": document,
        }

        if extra is not None:
            payload['extra'] = extra

        return payload

    def create_annotation_with_target_using_only_text_quote(self, url=None, prefix=None,
            exact=None, suffix=None, text=None, tags=None, tag_prefix=None,
            document=None, extra=None):
        """Call API with token and payload, create annotation (using only text quote)"""
        payload = self.make_annotation_payload_with_target_using_only_text_quote(url, prefix, exact, suffix,
                                                                                 text, tags, document, extra)
        try:
            r = self.post_annotation(payload)
        except BaseException as e:
            log.error(payload)
            log.exception(e)
            r = None  # if we get here someone probably ran the bookmarklet from firefox or the like
        return r

    def head_annotation(self, id):
        # used as a 'kind' way to look for deleted annotations
        headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8' }
        r = requests.head(self.api_url + '/annotations/' + id, headers=headers)
        return r

    def get_annotation(self, id):
        headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8' }
        r = requests.get(self.api_url + '/annotations/' + id, headers=headers)
        return r

    def post_annotation(self, payload):
        headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8' }
        data = json.dumps(payload, ensure_ascii=False)
        r = requests.post(self.api_url + '/annotations', headers=headers, data=data.encode('utf-8'))
        return r

    def patch_annotation(self, id, payload):
        headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8' }
        data = json.dumps(payload, ensure_ascii=False)
        r = requests.patch(self.api_url + '/annotations/' + id, headers=headers, data=data.encode('utf-8'))
        return r

    def delete_annotation(self, id):
        headers = {'Authorization': 'Bearer ' + self.token, 'Content-Type': 'application/json;charset=utf-8' }
        r = requests.delete(self.api_url + '/annotations/' + id, headers=headers)
        return r

    def search_all(self, params={}, max_results=None, stop_at=None):
        """Call search API with pagination, return rows """
        sort_by = params['sort'] if 'sort' in params else 'updated'
        if stop_at:
            if not isinstance(stop_at, str):
                raise TypeError('stop_at should be a string')

            dont_stop = (lambda r: r[sort_by] <= stop_at  # when ascending things less than stop are ok
                         if 'order' in params and params['order'] == 'asc'
                         else lambda r: r[sort_by] >= stop_at)

        if max_results:
            limit = 200 if 'limit' not in params else params['limit']  # FIXME hardcoded
            if max_results < limit:
                params['limit'] = max_results

        #sup_inf = max if params['order'] = 'asc' else min  # api defaults to desc
        # trust that rows[-1] works rather than potentially messsing stuff if max/min work differently
        nresults = 0
        while True:
            obj = self.search(params)
            rows = obj['rows']
            lr = len(rows)
            nresults += lr
            if lr is 0:
                return

            stop = None
            if max_results:
                if nresults >= max_results:
                    stop = max_results - nresults  + lr

            if stop_at:
                for row in rows[:stop]:
                    if dont_stop(row):
                        yield row
                    else:
                        return

            else:
                for row in rows[:stop]:
                    yield row

            if stop:
                return

            search_after = rows[-1][sort_by]
            params['search_after'] = search_after
            log.info(f'searching after {search_after}')

    def search_url(self, **params):
        return self.search_url_template.format(query=urlencode(params, True).replace('=','%3A'))  # = > :

    def query_url(self, **params):
        return self.query_url_template.format(query=urlencode(params, True))

    def search(self, params={}):
        """ Call search API, return a dict """
        if 'offset' not in params:
            params['offset'] = 0
        if 'limit' not in params or 'limit' in params and params['limit'] is None:
            params['limit'] = self.single_page_limit
        obj = self.authenticated_api_query(self.query_url(**params))
        return obj

    def batch_update(self, function, *ids, pretend=True):
        """ apply a function a set of annotations from their ids
            and patch the changes back to the remote
            pretend is set to True by default so that you can test the output
            before shooting yourself in the foot """
        # this is a friendly function that does sequential requests
        # if you want something e.g. with Async/deferred, don't use this

        if pretend:
            yield from ((a, function(a)) for id in ids
                        for a in (self.get_annotation(id).json(),))

        else:
            for id in ids:
                resp = self.get_annotation(id)
                blob = resp.json()
                updated = function(blob)
                yield blob, self.patch_annotation(id, updated)


class HypAnnoId(str):  # TODO derive from Identifer ...

    @property
    def iri(self):  # FIXME share or html ...
        return f'https://hyp.is/{self}'

    shareLink = iri

    @property
    def htmlLink(self):
        return f'https://hypothes.is/a/{self}'

    def compact(self, namespace_manager_ontcuries_localnamemanager_etc):
        """ given a compacting function, ball yourself up and put a bow on ya """
        # compact using rules from one of many, many implementations
        # of local <-> global name mapping
        # ids should retain the compact representation that they
        # were originally defined with
        # it seems simple enough to allow many other representations to
        # be generated by simply telling the identifer to ball itself
        # up according to the rules that it was passed ...
        # since the calling context clearly cares to have it shortened
        # according to those rules rathern than the original rules
        compactor = namespace_manager_ontcuries_localnamemanager_etc
        return compactor(self.iri)

    @property
    def curie(self):
        # the curie by convention
        return 'hyp:' + self  # FIXME need a better way ...

    def htmlCurie(self):
        return 'hypa:' + self  # FIXME not sure if correct ...

    @property
    def _asInstrumented(self):
        return Annotation(self, autofetch=False)  # FIXME distinguish/differentiate Anno vs HypAnno

    # FIXME decide on whether instrumentation should be a property or method
    # Given that in theory we can have an additional layer of not going to network
    # instantly by setting autofetch to False, the property approach might be ok
    # HOWEVER using a property like this makes it difficult to use asInstrumented
    # in Async/deferred to fetch a bunch of remote data in one shot
    # both variants are here, but I'm leaning toward the function -> autofetch
    # version, which will probably also be a better approach for OntId.asTerm as well...
    # patterns like Async(rate=lol)(deferred(id.asInstrumented)() for id in ids)
    # and Async(rate=lol)(deferred(AltAnnoImplementation)(id) for id in ids) both work
    # and make it easier to manage the balance between default and alternate remote stores

    def asInstrumented(self):
        return Annotation(self)  # FIXME distinguish/differentiate Anno vs HypAnno

    def __repr__(self):
        return f'{self.__class__.__name__}({self!r})'


class Annotation:
    """ a better object implementation for annotations needs to be merged with
        the HypothesisAnnotation implementation
    """

    immutable_fields = [
        'target',  # the target uri
        'exact',
    ]

    @classmethod
    def setup(cls, store=HypothesisUtils(username=username, group=group, token=api_token)):
        cls.store = store

    @classmethod
    def fromJson(cls, json):
        self = object.__new__(cls)  # skip the usual init process
        self._prior_versions = []  # FIXME as opposed @field.setter always returning a new object ...
        self._last_external = json  # the last thing we got from somewhere else
        self._json = json
        if 'id' in json:
            self.identifier = json['id']
        else:
            self.identifier = None

        return self

    def __init__(self, identifier, autofetch=True):
        self.identifier = identifier
        # FIXME the ability to have multiple stores that
        # cross reference eachother for robustness ...
        # 'this annotation can be found in x, y, and z'
        # stores under 'a, b, and c' identifiers respectively
        # for stuff like this having the store be able to
        # search by hash would be great ...
        if autofetch:
            self.data

    def _update_path(self, path, value):
        # TODO ...
        # a use case for adops ... ?
        current_value = getpath(self._json)
        if current_value != value:
            old = self._json
            new = deepcopy(self._json)
            setpath(new, value)
            self._prior_versions.append(old)
            self._json = new

    @property
    def data(self):
        # a bit overkill given how small most annotations are
        # but, no harm in matching the pattern we use everywhere
        # else, and setting autofetch = True in init so that
        # there is the option to disable for a potential
        # bulk request ...

        # FIXME autocache so that data always returns the
        # latest retrieved raw value? I think it should cache
        # and a new object should be returned for new versions
        # so that at least in this system the history of changes
        # can be preserved, even if the identifier stays the same
        # we apply the versioned doi approach where unqualified we
        # resolve to the usual landing page, and if you want a more
        # granular version then I think the updated datetime is the
        # simplest differentiator, primary key on id + updated ...
        # to allow the /{id}/version/{updated} pattern ...

        resp = self.store.get_annotation(self.identifier)
        self.headers = resp.headers
        self._json = resp.json()
        self._last_external = self._json
        return self._json

    def send(self, store=None):
        # FIXME post/patch/update send my data to the server please
        # send ..., yes send, ...
        # based on the remote service the annotation should just do
        # the right thing, and frankly an append only store just needs
        # to know the id
        if self.identifier is None:
            resp = self.store.post_annotation(self._json)
            blob = resp.json()
            self.identifier = blob['id']

        elif self._last_external != self._json:
            resp = self.store.patch_annotation(self._json)

        else:
            log.warning(f'Data for {self.identifier} has not changed, not sending.')

    def diff(self, other=None):
        raise NotImplementedError('since we are keeping prior versions ...')


class HypothesisAnnotation:
    """Encapsulate one row of a Hypothesis API search."""
    def __init__(self, row):
        if isinstance(row, HypothesisAnnotation):
            row = row._row

        self._row = row

    def _normalized(self):
        out = {}
        for k in dir(self):
            if not k.startswith('_'):
                v = getattr(self.__class__, k, None)  # need to walk up the mro
                if isinstance(v, property):
                    value = getattr(self, k)
                    if isinstance(value, GeneratorType):
                        value = list(value)

                    out[k] = value

        return out

    @property
    def id(self):
        return self._row['id']

    @property
    def user(self):
        return self._row['user'].replace('acct:','').replace('@hypothes.is','')

    @property
    def created(self):
        return self._row['created']

    @property
    def updated(self):
        return self._row['updated']

    @property
    def document(self):
        if 'document' in self._row:
            return self._row['document']
        else:
            return {}

    @property
    def filename(self):
        document = self.document
        if 'filename' in document:
            return document['filename']

    @property
    def doc_title(self):
        document = self.document
        if 'title' in document:
            title = document['title']
            if isinstance(title, list) and len(title):
                title = title[0]
        else:
            title = self.uri

        title = title.replace('"',"'")
        if not title:
            title = 'untitled'

        return title

    @property
    def links(self):
        document = self.document
        if 'link' in document:
            yield from document['link']

    @property
    def uri(self):
        if 'uri' in self._row:    # should it ever not?
            uri = self._row['uri']
        else:
             uri = "no uri field for %s" % self.id

        uri = uri.replace('https://via.hypothes.is/h/','').replace('https://via.hypothes.is/','')

        if uri.startswith('urn:x-pdf'):
            document = self.document
            for link in self.links:
                uri = link['href']
                if uri.encode('utf-8').startswith(b'urn:') == False:
                    break
            if uri.encode('utf-8').startswith(b'urn:') and 'filename' in document:
                uri = document['filename']

        return uri

    @property
    def tags(self):
        tags = []
        if 'tags' in self._row and self._row['tags'] is not None:
            tags = self._row['tags']
            if isinstance(tags, list):  # I find it hard to believe this is ever not true
                tags = [t.strip() for t in tags]
            else:
                raise BaseException('should never happen ...')

        return tags

    @property
    def text(self):
        text = ''
        if 'text' in self._row:
            text = self._row['text']

        return text

    @property
    def references(self):
        references = []
        if 'references' in self._row:
            references = self._row['references']

        return references

    @property
    def is_page_note(self):
        return self.type == 'pagenote'

    @property
    def type(self):
        if self.references:
            return 'reply'
        elif self.targets and any('selector' in t for t in self.targets):
            return 'annotation'
        else:
            return 'pagenote'

    @property
    def targets(self):
        # the spec says you can have multiple targets
        # the implementation only supports one at this time
        targets = []
        if 'target' in self._row:
            targets = self._row['target']

        return targets

    @property
    def selectors(self):
        for target in self.targets:
            if 'selector' in target:  # there are some that only have 'source'
                for selector in target['selector']:
                    yield selector

    def _selector_value(self, type, name):
        # obviously inefficient
        for selector in self.selectors:
            if 'type' in selector and selector['type'] == type:
                return selector[name]

    @property
    def prefix(self):
        return self._selector_value('TextQuoteSelector', 'prefix')

    @property
    def exact(self):
        return self._selector_value('TextQuoteSelector', 'exact')

    @property
    def suffix(self):
        return self._selector_value('TextQuoteSelector', 'suffix')

    @property
    def start(self):
        return self._selector_value('TextPositionSelector', 'start')

    @property
    def end(self):
        return self._selector_value('TextPositionSelector', 'end')

    @property
    def fragment_selector(self):
        return self._selector_value('FragmentSelector', 'value')

    @property
    def group(self): return self._row['group']

    @property
    def permissions(self): return {k:v for k,v in self._row['permissions'].items()}

    def __eq__(self, other):
        # text and tags can change, if exact changes then the id will also change
        return self.id == other.id and self.text == other.text and set(self.tags) == set(other.tags) and self.updated == other.updated

    def __hash__(self):
        return hash(self.id + self.text + self.updated)

    def __lt__(self, other):
        return self.updated < other.updated

    def __gt__(self, other):
        return not self.__lt__(other)


class iterclass(type):
    def __iter__(self):
        yield from self.objects.values()  # don't sort unless required

    @property
    def uri_tags(self):
        """ a dictionary all (processed) tags for a given uri """
        if not hasattr(self, '_uri_tags'):
            uri_tags = defaultdict(set)
            for obj in self.objects.values():  # do not use self here because the
                # sorting in __iter__ above can be extremely slow
                uri_tags[obj.uri].update(obj._tags)

            self._uri_tags = dict(uri_tags)  # FIXME this will go stale

        return self._uri_tags

    @property
    def uris(self):
        """ uris that have been annotated with tags from this workflow """
        if hasattr(self, 'namespace'):
            return set(uri for uri, tags in self.uri_tags.items()
                       if any(tag.startswith(self.prefix_ast)
                              for tag in tags))
        else:
            return set(self.uri_tags)

    @property
    def orphans(self):
        for id in self._orphanedReplies:
            yield self.byId(id)

# HypothesisHelper class customized to deal with replacing
#  exact, text, and tags based on its replies
#  also for augmenting the annotation with distinct fields
#  using annotation-text:exact or something like that... (currently using PROTCUR:annotation-exact which is super akward)
#  eg annotation-text:children to say exactly what the fields are when there needs to be more than one
#  it is possible to figure most of them out from their content but not always
class HypothesisHelper(metaclass=iterclass):  # a better HypothesisAnnotation
    """ A wrapper around sets of hypothes.is annotations
        with referential structure an pretty printing. """
    objects = {}  # TODO updates # NOTE: all child classes need their own copy of objects
    _tagIndex = {}
    _replies = {}
    reprReplies = True
    _embedded = False
    _done_loading = False
    _annos = {}
    _orphanedReplies = set()

    @classmethod
    def addAnno(cls, anno):
        return cls(anno, [anno])

    @classmethod
    def byId(cls, id_):
        try:
            return next(v for v in cls.objects.values()).getObjectById(id_)
        except StopIteration as e:
            raise Warning(f'{cls.__name__}.objects has not been populated with annotations yet!') from e

    @classmethod
    def byTags(cls, *tags):
        if cls._done_loading:  # TODO maybe better than done loading is 'consistent'?
            if not cls._tagIndex:
                log.debug('populating tags')
                # FIXME extremely inefficient on update
                # and we want this to update as replies appear
                # not all at once...
                # FIXME this does not update if new annos are added on the fly!
                [obj.populateTags() for obj in cls.objects.values()]

            return sorted(set.intersection(*(cls._tagIndex[tag] for tag in tags)))
        else:
            log.warning('attempted to search by tags before done loading')

    def populateTags(self):
        # FIXME need a way to evict old annos on update
        for tag in self.tags:
            if tag not in self._tagIndex:
                self._tagIndex[tag] = {self}
            else:
                self._tagIndex[tag].add(self)

            tset = self._tagIndex[tag]
            if tset not in self._remove_self_from:
                self._remove_self_from.append((tag, tset))

    def depopulateTags(self):
        """ remove object from the tag index on delete """
        log.debug(f'Removing {self._repr} from {len(self._remove_self_from)} tag sets')
        for tag, tset in self._remove_self_from:
            tset.remove(self)  # this should never error if everything is working correctly
            if not tset:  # remove unused tags from the index in depopulate
                self._tagIndex.pop(tag)

    @classmethod
    def byIri(cls, iri, prefix=False):
        norm_iri = norm(iri)
        for obj in cls.objects.values():
            norm_ouri = norm(obj.uri)
            if norm_ouri == norm_iri:
                yield obj
            elif prefix and norm_ouri.startswith(norm_iri):
                yield obj

    @classmethod
    def reset(cls):
        """ explicitly reset the class state removing _annos_list and _annos
            normally this should be called before the first time a program
            populates annotations so that any persistent state from another
            program is removed unfriendly if they need to coexist, but for that
            to actually work this whole thing needs a rewrite to have explicit
            representation of annotation groups """

        cls.objects = {}
        cls._tagIndex = {}
        cls._replies = {}
        cls.reprReplies = True
        cls._embedded = False
        cls._done_loading = False
        #HypothesisHelper._annos = {}  # DO NOT RESET THIS
        # the risk of staleness is worth it since we have
        # already worked through most of the possible issues
        # around things going stale for that
        # FIXME yes, yet another reason to switch to explicit
        # representation of in memory annotation stores

        for a in ('_annos_list',):
            if hasattr(cls, a):
                delattr(cls, a)

    def __new__(cls, anno, annos):
        if not hasattr(cls, '_annos_list'):
            cls._annos_list = annos
        elif cls._annos_list is not annos:  # FIXME STOP implement a real annos (SyncList) class FFS
            # hack to fix O(n ** 2) behavior or worse behavior
            # when readding the same set of annos over and over
            # I'm pretty sure that there is pathalogical behavior
            # hiding here because of the expectation that cls._annos_list is annos
            # for sync purposes ... sigh bad design coming back to haunt me again
            # having subclasses of HypothesisHelper act as singletons seems like
            # a good idea but eventually it will bite you
            sal = set(cls._annos_list)
            sa = set(annos)
            added = sa - sal
            removed = sal - sa
            new = [a for a in annos if a in added]
            cls._annos_list.extend(new)

            annos = cls._annos_list

        if hasattr(anno, 'deleted'):
            matches = [a for a in annos if a.id == anno.id]  # FIXME ick get rid of the list!
            for m in matches:
                cls._annos_list.remove(m)

            if anno.id in cls._annos:  # it is set to True by convetion
                cls._annos.pop(anno.id)  # insurance
            #else:
                #print("It's ok we already deleted", anno.id)
            if anno.id in cls.objects:
                obj = cls.objects.pop(anno.id)  # this is what we were missing
                obj.depopulateTags()
                #print('Found the sneek.', anno.id)
            return  # our job here is done

        if not cls._annos or len(cls._annos) < len(annos):  # much faster (as in O(n**2) -> O(1)) to populate once at the start
            # we should not need `if not a.deleted` because a should not be in annos
            cls._annos.update({a.id:a for a in annos})  # FIXME this fails on deletes...
            if len(cls._annos) != len(annos):
                logd.critical(f'it seems you have duplicate entries for annos: {len(cls._annos)} != {len(annos)}')
        try:
            self = cls.objects[anno.id]
            if self._updated == anno.updated:
                #printD(f'{self.id} already exists')
                return self
            else:
                #printD(f'{self.id} already exists but something has changed')
                cls._annos[anno.id] = anno  # update to the new anno version
                self.__init__(anno, annos)  # just updated the underlying refs no worries
                return self

        except KeyError:
            #printD(f'{anno.id} doesnt exist')
            return super().__new__(cls)

    def __init__(self, anno, annos):
        self._recursion_blocker = False
        self.annos = annos
        self.id = anno.id  # hardset this to prevent shenanigans
        self.objects[self.id] = self
        self._remove_self_from = []

        if self._tagIndex:
            # if tagIndex is not empty and we make it to __init__
            # then this anno helper has not been put into the tag index
            # FIXME stale annos in the tag index are likely an issue
            self.populateTags()

        if hasattr(self, '_uri_tags'):  # keep uri_tags in sync
            if anno.uri not in self._uri_tags:
                self._uri_tags[self.uri] = set()

            self._uri_tags[self.uri].update(self._tags)  # use _tags since tags can make many calls

        #if self.objects[self.id] is None:
            #printD('WAT', self.id)
        self.hasAstParent = False
        self.parent  # populate self._replies before the recursive call
        if len(self.objects) == len(annos):
            self.__class__._done_loading = True

    @property
    def classn(self):
        return self.__class__.__name__

    @property
    def _repr(self):
        return self.classn + f".byId('{self.id}')"

    @property
    def _anno(self):
        try:
            return self._annos[self.id]  # this way updateds to annos will propagate
        except KeyError as e:
            #from IPython import embed
            #embed()
            log.critical(str(self._tagIndex))
            raise e

    # protect the original annotation from modification
    @property
    def _permissions(self): return self._anno.permissions
    @property
    def _type(self): return self._anno.type
    @property
    def _exact(self): return self._anno.exact
    @property
    def _text(self): return self._anno.text
    @property
    def _tags(self): return self._anno.tags
    @property
    def _updated(self): return self._anno.updated  # amusing things happen if you use self._anno.tags instead...
    @property
    def references(self): return self._anno.references

    # we don't have any rules for how to modify these yet
    @property
    def exact(self): return self._exact
    @property
    def text(self): return self._text
    @property
    def tags(self): return self._tags
    @property
    def updated(self): return self._updated
    @property
    def uri(self): return self._anno.uri

    @classmethod
    def getAnnoById(cls, id_):
        try:
            return cls._annos[id_]
        except KeyError as e:
            #print('could not find', id_, shareLinkFromId(id_))
            return None

    def getObjectById(self, id_):
        try:
            return self.objects[id_]
        except KeyError as e:
            anno = self.getAnnoById(id_)
            if anno is None:
                #self.objects[id_] = None  # don't do this it breaks the type on objects
                #print('Problem in', self.shareLink)  # must come after self.objects[id_] = None else RecursionError
                if not self._recursion_blocker:
                    if self._type == 'reply':
                        #print('Orphaned reply', self.shareLink, f"{self.classn}.byId('{self.id}')")
                        self._orphanedReplies.add(self.id)
                    else:
                        logd.warning(f"Problem in {self.shareLink} {self.classn}.byId('{self.id}')")
                return None
            else:
                h = self.__class__(anno, self.annos)
                return h

    @property
    def shareLink(self):  # FIXME just look it up?!
        self._recursion_blocker = True
        if self.parent is not None:
            link = self.parent.shareLink
            # call link before unset recursion to prevent cases
            # where an intermediate parent was deleted
            self._recursion_blocker = False
            return link
        else:
            self._recursion_blocker = False
            return shareLinkFromId(self.id)

    @property
    def htmlLink(self):
        return self._anno._row['links']['html']

    @property
    def parent(self):
        if not self.references:
            return None
        else:
            for parent_id in self.references[::-1]:  # go backward to get the direct parent first, slower for shareLink but ok
                parent = self.getObjectById(parent_id)
                if parent is not None:
                    if parent.id not in self._replies:
                        self._replies[parent.id] = set()
                    self._replies[parent.id].add(self)
                    return parent
                else:
                    #printD(f"Parent gone for {self.__class__.__name__}.byId('{self.id}'}")
                    pass

    @property
    def replies(self):
        # for the record, the naieve implementation of this
        # looping over annos everytime is 3 orders of magnitude slower
        if self._done_loading:
            if self.id not in self._replies:
                self._replies[self.id] = set()
            return self._replies[self.id]  # we use self.id here instead of self to avoid recursion on __eq__
        else:
            log.warning('Not done loading annos, you will be missing references!')
            return set()


    def __eq__(self, other):
        return (type(self) == type(other) and
                self.id == other.id and
                self.text == other.text and
                set(self.tags) == set(other.tags) and
                self.updated == other.updated)

    def __hash__(self):
        return hash(self.__class__.__name__ + self.id)

    def __lt__(self, other):
        return self.updated < other.updated

    def __gt__(self, other):
        return not self.__lt__(other)

    @property
    def _python__repr__(self):
        return f"{self.__class__.__name__}.byId('{self.id}')"

    def __repr__(self, depth=0, format__repr__for_children='', html=False, number='*'):
        start = '|' if depth else ''
        SPACE = '&nbsp;' if html else ' '
        t = SPACE * 4 * depth + start

        parent_id =  f"\n{t}parent_id:    {self.parent.id} {self.parent._python__repr__}" if self.parent else ''
        exact_text = f'\n{t}exact:        {self.exact}' if self.exact else ''

        text_align = 'text:         '
        lp = f'\n{t}'
        text_line = lp + ' ' * len(text_align)
        text_text = lp + text_align + self.text.replace('\n', text_line) if self.text else ''
        tag_text =   f'\n{t}tags:         {self.tags}' if self.tags else ''

        replies = ''.join(r.__repr__(depth + 1) for r in self.replies)
        rep_ids = f'\n{t}replies:      ' + ' '.join(r._python__repr__ for r in self.replies)
        replies_text = (f'\n{t}replies:{replies}' if self.reprReplies else rep_ids) if replies else ''
        link = self.shareLink
        if html: link = atag(link, link)
        startn = '\n' if not isinstance(number, int) or number > 1 else ''
        return (f'{startn}{t.replace("|","")}{number:-<20}'
                f"\n{t}{self.__class__.__name__ + ':':<14}{link} {self._python__repr__}"
                f'\n{t}user:         {self._anno.user}'
                f'{parent_id}'
                f'{exact_text}'
                f'{text_text}'
                f'{tag_text}'
                f'{replies_text}'
                f'{format__repr__for_children}'
                f'\n{t}{"":_<20}')




# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from hyputils.memex.util import markdown


class TestRender(object):
    def test_it_renders_markdown(self):
        actual = markdown.render("_emphasis_ **bold**")
        assert "<p><em>emphasis</em> <strong>bold</strong></p>\n" == actual

    def test_it_ignores_math_block(self):
        actual = markdown.render("$$1 + 1 = 2$$")
        assert "<p>$$1 + 1 = 2$$</p>\n" == actual

    def test_it_ignores_inline_match(self):
        actual = markdown.render(r"Foobar \(1 + 1 = 2\)")
        assert "<p>Foobar \\(1 + 1 = 2\\)</p>\n" == actual

    def test_it_sanitizes_the_output(self, markdown_render, sanitize):
        markdown.render("foobar")
        sanitize.assert_called_once_with(markdown_render.return_value)

    @pytest.fixture
    def markdown_render(self, patch):
        return patch("hyputils.memex.util.markdown.markdown")

    @pytest.fixture
    def sanitize(self, patch):
        return patch("hyputils.memex.util.markdown.sanitize")


class TestSanitize(object):
    @pytest.mark.parametrize(
        "text,expected",
        [
            (
                '<a href="https://example.org">example</a>',
                '<a href="https://example.org" rel="nofollow noopener" target="_blank">example</a>',
            ),
            # Don't add rel and target attrs to mailto: links
            ('<a href="mailto:foo@example.net">example</a>', None),
            ('<a title="foobar">example</a>', None),
            (
                '<a href="https://example.org" rel="nofollow noopener" target="_blank" title="foobar">example</a>',
                None,
            ),
            ("<blockquote>Foobar</blockquote>", None),
            ("<code>foobar</code>", None),
            ("<em>foobar</em>", None),
            ("<hr>", None),
            ("<h1>foobar</h1>", None),
            ("<h2>foobar</h2>", None),
            ("<h3>foobar</h3>", None),
            ("<h4>foobar</h4>", None),
            ("<h5>foobar</h5>", None),
            ("<h6>foobar</h6>", None),
            ('<img src="http://example.com/img.jpg">', None),
            ('<img src="/img.jpg">', None),
            ('<img alt="foobar" src="/img.jpg">', None),
            ('<img src="/img.jpg" title="foobar">', None),
            ('<img alt="hello" src="/img.jpg" title="foobar">', None),
            ("<ol><li>foobar</li></ol>", None),
            ("<p>foobar</p>", None),
            ("<pre>foobar</pre>", None),
            ("<strong>foobar</strong>", None),
            ("<ul><li>foobar</li></ul>", None),
        ],
    )
    def test_it_allows_markdown_html(self, text, expected):
        if expected is None:
            expected = text

        assert markdown.sanitize(text) == expected

    @pytest.mark.parametrize(
        "text,expected",
        [
            ("<script>evil()</script>", "&lt;script&gt;evil()&lt;/script&gt;"),
            (
                '<a href="#" onclick="evil()">foobar</a>',
                '<a href="#" rel="nofollow noopener" target="_blank">foobar</a>',
            ),
            (
                '<a href="#" onclick=evil()>foobar</a>',
                '<a href="#" rel="nofollow noopener" target="_blank">foobar</a>',
            ),
            ("<a href=\"javascript:alert('evil')\">foobar</a>", "<a>foobar</a>"),
            ('<img src="/evil.jpg" onclick="evil()">', '<img src="/evil.jpg">'),
            ("<img src=\"javascript:alert('evil')\">", "<img>"),
        ],
    )
    def test_it_escapes_evil_html(self, text, expected):
        assert markdown.sanitize(text) == expected

    def test_it_adds_target_blank_and_rel_nofollow_to_links(self):
        actual = markdown.sanitize('<a href="https://example.org">Hello</a>')
        expected = '<a href="https://example.org" rel="nofollow noopener" target="_blank">Hello</a>'

        assert actual == expected



# -*- coding: utf-8 -*-

"""Reusable mixins for SQLAlchemy declarative models."""

from __future__ import unicode_literals

import datetime

import sqlalchemy as sa


class Timestamps(object):
    created = sa.Column(
        sa.DateTime,
        default=datetime.datetime.utcnow,
        server_default=sa.func.now(),
        nullable=False,
    )
    updated = sa.Column(
        sa.DateTime,
        server_default=sa.func.now(),
        default=datetime.datetime.utcnow,
        onupdate=datetime.datetime.utcnow,
        nullable=False,
    )



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import factory

from hyputils.memex import models

from .base import ModelFactory


def unique_username(obj):
    """
    Return a unique username for a test user.

    Return a randomly generated username that is guaranteed to be unique within
    a given test run. Uniqueness is necessary because usernames must be unique
    in the DB and generating random usernames in a not-guaranteed-to-be-unique
    way would result in intermittent database errors when running the tests.

    """
    return obj.non_unique_username + "__" + obj.count


def unique_email(obj):
    """
    Return a unique email for a test user.

    Return a randomly generated email that is guaranteed to be unique within
    a given test run. Uniqueness is necessary because emails must be unique
    in the DB and generating random emails in a not-guaranteed-to-be-unique
    way would result in intermittent database errors when running the tests.

    """
    return obj.username + "@" + obj.email_domain


class User(ModelFactory):
    class Meta:
        model = models.User

    class Params:
        # A count that's appended to non-unique usernames to make them unique.
        count = factory.Sequence(lambda n: "%d" % n)
        # The non-unique part of the generated username.
        non_unique_username = factory.Faker("user_name")
        # The domain (following ``@``) part of the generated email address.
        email_domain = factory.Faker("free_email_domain")

    authority = "example.com"
    username = factory.LazyAttribute(unique_username)
    email = factory.LazyAttribute(unique_email)
    registered_date = factory.Faker("date_time_this_decade")



# -*- coding: utf-8 -*-
"""Helpers for the Python 2 to Python 3 transition."""
from __future__ import unicode_literals

import sys

__all__ = (
    "PY2",
    "text_type",
    "string_types",
    "configparser",
    "urlparse",
    "url_quote",
    "url_quote_plus",
    "url_unquote",
    "url_unquote_plus",
    "StringIO",
)

PY2 = sys.version_info[0] == 2

if not PY2:
    text_type = str
    string_types = (str,)
    xrange = range
    unichr = chr
else:
    text_type = unicode  # noqa
    string_types = (str, unicode)  # noqa
    xrange = xrange
    unichr = unichr

try:
    import ConfigParser as configparser
except ImportError:
    import configparser

try:
    from urllib import parse as urlparse

    url_quote = urlparse.quote
    url_quote_plus = urlparse.quote_plus
    url_unquote = urlparse.unquote
    url_unquote_plus = urlparse.unquote_plus
except ImportError:
    import urllib
    import urlparse

    url_quote = urllib.quote
    url_quote_plus = urllib.quote_plus
    url_unquote = urllib.unquote
    url_unquote_plus = urllib.unquote_plus

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


# native() function adapted from Pyramid:
# https://github.com/Pylons/pyramid/blob/a851d05f76edc6bc6bf65269c20eeba7fe726ade/pyramid/compat.py#L78-L95
if PY2:

    def native(s, encoding="latin-1", errors="strict"):
        """
        Return the given string as a Python 2 native string (a byte string).

        If the given string is a unicode string then return it as a byte
        string (Latin 1 encoded by default).

        If the given string is already a byte string (in any encoding) just
        return it unmodified.

        Latin 1 encoding is used by default because that's the encoding used
        for "native" strings in PEP-3333 (the WSGI spec).

        """
        if isinstance(s, unicode):  # noqa
            return s.encode(encoding, errors)
        return s


else:

    def native(s, encoding="latin-1", errors="strict"):
        """
        Return the given string as a Python 3 native string (a unicode string).

        If the given string is a byte string then return it decoded to unicode
        (using Latin 1 by default).

        If the given string is already a unicode string then just return it
        unmodified.

        Latin 1 encoding is used by default because that's the encoding used
        for "native" strings in PEP-3333 (the WSGI spec).

        """
        if isinstance(s, str):
            return s
        return str(s, encoding, errors)



# -*- coding: utf-8 -*-
# pylint: disable=no-self-use
"""
The `conftest` module is automatically loaded by pytest and serves as a place
to put fixture functions that are useful application-wide.
"""
from __future__ import unicode_literals

import functools
import os

import mock
import pytest

import sqlalchemy
from sqlalchemy.orm import sessionmaker
from webob.multidict import MultiDict

from hyputils.memex import db
from hyputils.memex import models
from hyputils.memex.settings import database_url
from hyputils.memex._compat import text_type

TEST_AUTHORITY = "example.com"
TEST_DATABASE_URL = database_url(
    os.environ.get("TEST_DATABASE_URL", "postgresql://postgres@localhost/htest")
)

Session = sessionmaker()


class DummySession(object):

    """
    A dummy database session.
    """

    def __init__(self):
        self.added = []
        self.deleted = []
        self.flushed = False

    def add(self, obj):
        self.added.append(obj)

    def delete(self, obj):
        self.deleted.append(obj)

    def flush(self):
        self.flushed = True


class DummyRequest(object):
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


def autopatcher(request, target, **kwargs):
    """Patch and cleanup automatically. Wraps :py:func:`mock.patch`."""
    options = {"autospec": True}
    options.update(kwargs)
    patcher = mock.patch(target, **options)
    obj = patcher.start()
    request.addfinalizer(patcher.stop)
    return obj


@pytest.fixture(scope="session")
def db_engine():
    """Set up the database connection and create tables."""
    engine = sqlalchemy.create_engine(TEST_DATABASE_URL)
    db.init(engine, should_create=True, should_drop=True, authority=TEST_AUTHORITY)
    return engine


@pytest.fixture
def default_organization(db_session):
    return models.Organization.default(db_session)


@pytest.fixture
def db_session(db_engine):
    """
    Prepare the SQLAlchemy session object.

    We enable fast repeatable database tests by setting up the database only
    once per session (see :func:`db_engine`) and then wrapping each test
    function in a transaction that is rolled back.

    Additionally, we set a SAVEPOINT before entering the test, and if we
    detect that the test has committed (i.e. released the savepoint) we
    immediately open another. This has the effect of preventing test code from
    committing the outer transaction.
    """
    conn = db_engine.connect()
    trans = conn.begin()
    session = Session(bind=conn)
    session.begin_nested()

    @sqlalchemy.event.listens_for(session, "after_transaction_end")
    def restart_savepoint(session, transaction):
        if transaction.nested and not transaction._parent.nested:
            session.begin_nested()

    try:
        yield session
    finally:
        session.close()
        trans.rollback()
        conn.close()


@pytest.fixture
def factories(db_session):
    from ..common import factories

    factories.set_session(db_session)
    yield factories
    factories.set_session(None)


@pytest.fixture
def fake_db_session():
    return DummySession()


@pytest.fixture
def matchers():
    from ..common import matchers

    return matchers


@pytest.fixture
def patch(request):
    return functools.partial(autopatcher, request)


@pytest.fixture
def pyramid_request(db_session):
    """Dummy Pyramid request object."""
    request = DummyRequest(db=db_session)
    request.authenticated_userid = 'THIS IS NOT ACTULLY A USERID LOL'
    request.default_authority = text_type(TEST_AUTHORITY)
    request.create_form = mock.Mock()
    request.matched_route = mock.Mock()
    request.is_xhr = False
    request.params = MultiDict()
    request.GET = request.params
    request.POST = request.params
    request.user = None
    return request



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
from sqlalchemy import exc
from hyputils.memex.security import security

from hyputils.memex import models


class TestUserModelDataConstraints(object):
    """Unit tests for :py:module:`h.models.User` data integrity constraints"""

    def test_cannot_create_dot_variant_of_user(self, db_session):
        fred = models.User(
            authority="example.com", username="fredbloggs", email="fred@example.com"
        )
        fred2 = models.User(
            authority="example.com", username="fred.bloggs", email="fred@example.org"
        )

        db_session.add(fred)
        db_session.add(fred2)
        with pytest.raises(exc.IntegrityError):
            db_session.flush()

    def test_cannot_create_case_variant_of_user(self, db_session):
        bob = models.User(
            authority="example.com", username="BobJones", email="bob@example.com"
        )
        bob2 = models.User(
            authority="example.com", username="bobjones", email="bob@example.org"
        )

        db_session.add(bob)
        db_session.add(bob2)
        with pytest.raises(exc.IntegrityError):
            db_session.flush()

    def test_filtering_by_username_matches_dot_variant_of_user(self, db_session):
        fred = models.User(
            authority="example.com", username="fredbloggs", email="fred@example.com"
        )
        db_session.add(fred)
        db_session.flush()

        result = db_session.query(models.User).filter_by(username="fred.bloggs").one()

        assert result == fred

    def test_filtering_by_username_matches_case_variant_of_user(self, db_session):
        fred = models.User(
            authority="example.com", username="fredbloggs", email="fred@example.com"
        )
        db_session.add(fred)
        db_session.flush()

        result = db_session.query(models.User).filter_by(username="FredBloggs").one()

        assert result == fred

    def test_userid_derived_from_username_and_authority(self):
        fred = models.User(
            authority="example.net", username="fredbloggs", email="fred@example.com"
        )

        assert fred.userid == "acct:fredbloggs@example.net"

    def test_cannot_create_user_with_too_short_username(self):
        with pytest.raises(ValueError):
            models.User(username="aa")

    def test_cannot_create_user_with_too_long_username(self):
        with pytest.raises(ValueError):
            models.User(username="1234567890123456789012345678901")

    def test_cannot_create_user_with_invalid_chars(self):
        with pytest.raises(ValueError):
            models.User(username="foo-bar")

    def test_cannot_create_user_with_too_long_email(self):
        with pytest.raises(ValueError):
            models.User(email="bob@b" + "o" * 100 + "b.com")

    def test_can_create_user_with_null_email(self):
        models.User(email=None)

    def test_can_change_email_to_null(self):
        user = models.User(email="bob@bob.com")

        user.email = None

    def test_cannot_create_two_users_with_same_non_null_email_and_authority(
        self, db_session, factories
    ):
        factories.User(email="bob@bob.com", authority="hypothes.is")
        factories.User(email="bob@bob.com", authority="hypothes.is")

        with pytest.raises(
            exc.IntegrityError,
            match='duplicate key value violates unique constraint "uq__user__email"',
        ):
            db_session.flush()

    def test_can_create_two_users_with_same_null_email_and_authority(
        self, db_session, factories
    ):
        factories.User(email=None, authority="hypothes.is")
        factories.User(email=None, authority="hypothes.is")

        db_session.flush()


class TestUserModelUserId(object):
    def test_userid_equals_query(self, db_session):
        fred = models.User(
            authority="example.net", username="fredbloggs", email="fred@example.com"
        )
        db_session.add(fred)
        db_session.flush()

        result = (
            db_session.query(models.User)
            .filter_by(userid="acct:fredbloggs@example.net")
            .one()
        )

        assert result == fred

    def test_userid_equals_query_with_invalid_userid(self, db_session):
        # This is to ensure that we don't expose the ValueError that could
        # potentially be thrown by split_user.

        result = (
            db_session.query(models.User)
            .filter_by(userid="fredbloggsexample.net")
            .all()
        )

        assert result == []

    def test_userid_in_query(self, db_session):
        fred = models.User(
            authority="example.net", username="fredbloggs", email="fred@example.net"
        )
        alice = models.User(
            authority="foobar.com", username="alicewrites", email="alice@foobar.com"
        )
        db_session.add_all([fred, alice])
        db_session.flush()

        result = (
            db_session.query(models.User)
            .filter(
                models.User.userid.in_(
                    [
                        "acct:fredbloggs@example.net",
                        "acct:alicewrites@foobar.com",
                        "acct:missing@bla.org",
                    ]
                )
            )
            .all()
        )

        assert len(result) == 2
        assert fred in result
        assert alice in result

    def test_userid_in_query_with_invalid_userid_mixed_in(self, db_session):
        # This is to ensure that we don't expose the ValueError that could
        # potentially be thrown by split_user.

        fred = models.User(
            authority="example.net", username="fredbloggs", email="fred@example.com"
        )
        db_session.add(fred)
        db_session.flush()

        result = (
            db_session.query(models.User)
            .filter(models.User.userid.in_(["acct:fredbloggs@example.net", "invalid"]))
            .all()
        )

        assert len(result) == 1
        assert fred in result

    def test_userid_in_query_with_only_invalid_userid(self, db_session):
        # This is to ensure that we don't expose the ValueError that could
        # potentially be thrown by split_user.

        result = (
            db_session.query(models.User)
            .filter(models.User.userid.in_(["fredbloggsexample.net"]))
            .all()
        )

        assert result == []


class TestUserGetByEmail(object):
    def test_it_returns_a_user(self, db_session, users):
        user = users["meredith"]
        actual = models.User.get_by_email(db_session, user.email, user.authority)
        assert actual == user

    def test_it_filters_by_email(self, db_session, users):
        authority = "example.com"
        email = "bogus@msn.com"

        actual = models.User.get_by_email(db_session, email, authority)
        assert actual is None

    def test_it_filters_email_case_insensitive(self, db_session, users):
        user = users["emily"]
        mixed_email = "eMiLy@mSn.com"

        actual = models.User.get_by_email(db_session, mixed_email, user.authority)
        assert actual == user

    def test_it_filters_by_authority(self, db_session, users):
        user = users["norma"]

        actual = models.User.get_by_email(db_session, user.email, "example.com")
        assert actual is None

    def test_you_cannot_get_users_with_no_emails(self, db_session):
        assert not models.User.get_by_email(db_session, None, "example.com")

    @pytest.fixture
    def users(self, db_session, factories):
        users = {
            "emily": factories.User(
                username="emily", email="emily@msn.com", authority="example.com"
            ),
            "norma": factories.User(
                username="norma", email="norma@foo.org", authority="foo.org"
            ),
            "meredith": factories.User(
                username="meredith", email="meredith@gmail.com", authority="example.com"
            ),
            "bob": factories.User(username="bob", email=None, authority="example.com"),
        }
        db_session.flush()
        return users


class TestUserGetByUsername(object):
    def test_it_returns_a_user(self, db_session, users):
        user = users["meredith"]

        actual = models.User.get_by_username(db_session, user.username, user.authority)
        assert actual == user

    def test_it_filters_by_username(self, db_session):
        authority = "example.com"
        username = "bogus"

        actual = models.User.get_by_username(db_session, username, authority)
        assert actual is None

    def test_it_filters_by_authority(self, db_session, users):
        user = users["norma"]

        actual = models.User.get_by_username(db_session, user.username, "example.com")
        assert actual is None

    @pytest.fixture
    def users(self, db_session, factories):
        users = {
            "emily": factories.User(username="emily", authority="example.com"),
            "norma": factories.User(username="norma", authority="foo.org"),
            "meredith": factories.User(username="meredith", authority="example.com"),
        }
        db_session.flush()
        return users


class TestUserACL(object):
    def test_auth_client_with_matching_authority_may_update_user(
        self, user, authz_policy
    ):
        user.authority = "weewhack.com"

        assert authz_policy.permits(
            user, ["flip", "client_authority:weewhack.com"], "update"
        )

    def test_auth_client_without_matching_authority_may_not_update_user(
        self, user, authz_policy
    ):
        user.authority = "weewhack.com"

        assert not authz_policy.permits(
            user, ["flip", "client_authority:2weewhack.com"], "update"
        )

    def test_user_with_authority_may_not_update_user(self, user, authz_policy):
        user.authority = "fabuloso.biz"

        assert not authz_policy.permits(
            user, ["flip", "authority:fabuloso.biz"], "update"
        )

    @pytest.fixture
    def authz_policy(self):
        return security.ACLAuthorizationPolicy()

    @pytest.fixture
    def user(self):
        user = models.User(username="filip", authority="example.com")
        return user



#!/usr/bin/env python3.6

import os
import asyncio
import ssl
import uuid
import json
from os import environ
from socket import socketpair
from threading import Thread

import certifi
import websockets


class Handler:
    def __init__(self, filter_handlers):
        self.filter_handlers = filter_handlers  # list of filterHandlers that should be run on every message

    def process(self, message):
        if message['type'] == 'annotation-notification':
            for fh in self.filter_handlers:
                fh(message)
        else:
            print('NOT ANNOTATION')
            print(message)


class preFilter:
    """ Create a filter that will run on the hypothes.is server
        Make group empty to default to allow all groups the authed user
        is a member of in the hypothes.is system.
    """
    def __init__(self, groups=[], users=[], uris=[], tags=[],
                 create=True, update=True, delete=True,
                 match_policy='include_any'):

        self.create = create
        self.update = update
        self.delete = delete

        #include_all include_any
        self.match_policy = match_policy

        self.groups = groups
        self.users = users
        self.uris = uris  # NOTE: uri filters must be exact :(
        self.tags = tags

        self.clause_map = [
            ('/group', self.groups),  # __world__
            ('/user', self.users),
            ('/uri', self.uris),
            ('/tags', self.tags),
        ]

    def _make_clauses(self):
        clauses = []
        for field, value in self.clause_map:
            if value:
                clauses.append(
                    {'field':field,
                     'case_sensitive':True,
                     'operator':'one_of',
                     'options':{},
                     'value':value,
                    }
                )

        return clauses

    def export(self):
        output = {
            'filter':{
                'actions':{
                    'create':self.create,
                    'update':self.update,
                    'delete':self.delete,
                },
                'match_policy':self.match_policy,
                'clauses':self._make_clauses(),
            },
        }

        return output


def _ssl_context(verify=True):
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    if not verify:
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
    return ssl_context


async def setup_connection(websocket):
    message = {'messageType': 'client_id',
               'value': str(uuid.uuid4()),}
    print('SETUP MESSAGE', message)
    await websocket.send(json.dumps(message))


async def setup_filters(websocket, filters):
    print('SETUP FILTERS\n', json.dumps(filters, indent=2))
    await websocket.send(json.dumps(filters))


async def process_messages(websocket, handler):
    while True:
        response = await websocket.recv()
        try:
            msg = json.loads(response)
        except ValueError:
            pass
        if msg:
            handler.process(msg)
        else:
            pass


class ExitLoop(Exception):
    pass


async def listen_for_exit(reader):
    # the managing process will send a message on exit
    msg = await reader.readline()
    raise ExitLoop(msg.decode())


async def process_or_exit(websock, handler, exit_reader):
    process_task = asyncio.ensure_future(process_messages(websock, handler))
    exit_task = asyncio.ensure_future(listen_for_exit(exit_reader))
    done, pending = await asyncio.wait([process_task, exit_task],
                                       return_when=asyncio.FIRST_EXCEPTION)
    future = done.pop()
    for task in pending:
        task.cancel()
    raise future.exception()


def setup_websocket(api_token, filters, filter_handlers,
                    websocket_endpoint='wss://hypothes.is/ws',
                    extra_headers=None):
    if extra_headers is None:
        extra_headers = {}

    rsock, wsock = socketpair()

    def exit_loop():
        try:
            # stop the current await
            wsock.send(b'Parent processes sent exit\n')
            # close the socket and make sure we don't start again
            # or more simply, to avoid leaking resources
            wsock.close()
        except OSError:
            pass  # socket was already closed

    async def ws_loop(loop):
        #websocket_endpoint = 'wss://hypothes.is/ws'
        #filter_handlers = getFilterHandlers()
        handler = Handler(filter_handlers)
        ssl_context = _ssl_context(verify=True)

        headers = {'Authorization': 'Bearer ' + api_token}
        extra_headers.update(headers)

        exit_reader, _writer = await asyncio.open_connection(sock=rsock, loop=loop)

        while True:  # for insurance could also test on closed wsock
            print('WE SHOULD GET HERE')
            try:
                async with websockets.connect(websocket_endpoint,
                                              ssl=ssl_context,
                                              extra_headers=extra_headers) as ws:
                    await setup_connection(ws)
                    print(f'websocket connected to {websocket_endpoint}')
                    await setup_filters(ws, filters)
                    print('subscribed')
                    await process_or_exit(ws, handler, exit_reader)
            except ExitLoop as e:  # for whatever reason the await proceess or exit doesn't work here :/
                print(e)
                break
            except KeyboardInterrupt as e:
                break
            except (websockets.exceptions.ConnectionClosed, ConnectionResetError) as e:
                pass

        _writer.close()  # prevents ResourceWarning

    return ws_loop, exit_loop


class AnnotationStream:
    def __init__(self, annos, prefilter, *handler_classes, memoizer=None):
        from .hypothesis import api_token
        self.api_token = api_token
        self.annos = annos
        self.filters = prefilter
        self.filter_handlers = [handler(self.annos, memoizer=memoizer) for handler in handler_classes]

    @staticmethod
    def loop_target(loop, ws_loop):
        asyncio.set_event_loop(loop)
        loop.run_until_complete(ws_loop(loop))

    def __call__(self):
        loop = asyncio.get_event_loop()
        ws_loop, exit_loop = setup_websocket(self.api_token, self.filters, self.filter_handlers)
        stream_thread = Thread(target=self.loop_target, args=(loop, ws_loop))
        return stream_thread, exit_loop


def main():
    from handlers import printHandler, websocketServerHandler
    loop = asyncio.get_event_loop()

    subscribed = {}
    def send_message(d):
        for send in subscribed.values():
            send(json.dumps(d).encode())

    wssh = websocketServerHandler(send_message)

    async def incoming_handler(websocket, path):
        try:
            await websocket.recv()  # do nothing except allow us to detect unsubscribe
        except websockets.exceptions.ConnectionClosed as e:
            pass  # working as expected

    async def outgoing_handler(websocket, path, reader):
        while True:
            message = await reader.readline()
            await websocket.send(message.decode())

    async def conn_handler(websocket, path, reader):
        i_task = asyncio.ensure_future(incoming_handler(websocket, path))
        o_task = asyncio.ensure_future(outgoing_handler(websocket, path, reader))
        done, pending = await asyncio.wait([i_task, o_task], return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()

    async def subscribe(websocket, path):
        name = await websocket.recv()  # this is not needed...
        print(f"< {name}")
        greeting = json.dumps(f"Hello {name}! You are now subscribed to cat facts!{{}}"
                              f"{list(subscribed)} are also subscribed to cat facts!")
        greeting = greeting.format('\n')
        rsock, wsock = socketpair()
        reader, writer = await asyncio.open_connection(sock=rsock, loop=loop)
        for send_something in subscribed.values():
            msg = json.dumps(f'{name} also subscribed to cat facts!').encode()
            send_something(msg)

        def send(bytes_, s=wsock.send):
            s(bytes_)
            s(b'\n')

        subscribed[name] = send  # _very_ FIXME NOTE this is how we know where to route all our messages
        await websocket.send(greeting)
        print(f"> {greeting}")
        # we now wait here for something else to happen, in this case
        # either there is a subscription or an unsubscription
        await conn_handler(websocket, path, reader)  # when this completes the connection is closed
        subscribed.pop(name)
        for send_something in subscribed.values():
            msg = json.dumps(f'{name} unsubscribed from cat facts!').encode()
            send_something(msg)

    start_server = websockets.serve(subscribe, 'localhost', 5050)
    loop.run_until_complete(start_server)  # TODO need this wrapped so that loop can be passed in

    api_token = environ.get('HYP_API_TOKEN', 'TOKEN')
    groups = environ.get('HYP_GROUPS', '__world__').split(' ')
    filters = preFilter(groups=groups).export()
    filter_handlers = [printHandler(), wssh]
    print(groups)
    ws_loop, exit_loop = setup_websocket(api_token, filters, filter_handlers)

    loop.run_until_complete(ws_loop(loop))

if __name__ == '__main__':
    main()




# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import datetime

import mock
import pytest
import sqlalchemy as sa

from hyputils.memex import models
from hyputils.memex.models import document


class TestDocumentFindByURIs(object):
    def test_with_one_matching_Document(self, db_session):
        # One Document with a non-matching DocumentURI pointing to it.
        # find_by_uris() should not return this Document.
        document1 = document.Document()
        uri1 = "https://de.wikipedia.org/wiki/Hauptseite"
        document1.document_uris.append(document.DocumentURI(claimant=uri1, uri=uri1))

        # A second Document with one matching and one non-matching DocumentURI
        # pointing to it. find_by_uris() should return this Document.
        document2 = document.Document()
        uri2 = "https://en.wikipedia.org/wiki/Main_Page"
        document2.document_uris.append(document.DocumentURI(claimant=uri2, uri=uri2))
        uri3 = "https://en.wikipedia.org"
        document2.document_uris.append(document.DocumentURI(claimant=uri3, uri=uri2))

        db_session.add_all([document1, document2])
        db_session.flush()

        actual = document.Document.find_by_uris(
            db_session,
            [
                "https://en.wikipedia.org/wiki/Main_Page",
                "https://m.en.wikipedia.org/wiki/Main_Page",
            ],
        )

        assert actual.count() == 1
        assert actual.first() == document2

    def test_no_matches(self, db_session):
        document_ = document.Document()
        document_.document_uris.append(
            document.DocumentURI(
                claimant="https://en.wikipedia.org/wiki/Main_Page",
                uri="https://en.wikipedia.org/wiki/Main_Page",
            )
        )
        db_session.add(document_)
        db_session.flush()

        actual = document.Document.find_by_uris(
            db_session, ["https://de.wikipedia.org/wiki/Hauptseite"]
        )
        assert actual.count() == 0


class TestDocumentFindOrCreateByURIs(object):
    def test_with_one_existing_Document(self, db_session):
        """
        When there's one matching Document it should return that Document.

        When searching with two URIs that match two DocumentURIs that both
        point to the same Document, it should return that Document.

        """
        document_ = document.Document()
        docuri1 = document.DocumentURI(
            claimant="https://en.wikipedia.org/wiki/Main_Page",
            uri="https://en.wikipedia.org/wiki/Main_Page",
            document=document_,
        )
        docuri2 = document.DocumentURI(
            claimant="https://en.wikipedia.org/wiki/http/en.m.wikipedia.org/wiki/Main_Page",
            uri="https://en.wikipedia.org/wiki/Main_Page",
            document=document_,
        )

        db_session.add(docuri1)
        db_session.add(docuri2)
        db_session.flush()

        actual = document.Document.find_or_create_by_uris(
            db_session,
            "https://en.wikipedia.org/wiki/Main_Page",
            [
                "https://en.wikipedia.org/wiki/http/en.m.wikipedia.org/wiki/Main_Page",
                "https://m.en.wikipedia.org/wiki/Main_Page",
            ],
        )

        assert actual.count() == 1
        assert actual.first() == document_

    def test_with_no_existing_documents(self, db_session):
        """When there are no matching Documents it creates and returns one."""
        document_ = document.Document()
        docuri = document.DocumentURI(
            claimant="https://en.wikipedia.org/wiki/Main_Page",
            uri="https://en.wikipedia.org/wiki/Main_Page",
            document=document_,
        )

        db_session.add(docuri)
        db_session.flush()

        documents = document.Document.find_or_create_by_uris(
            db_session,
            "https://en.wikipedia.org/wiki/Pluto",
            ["https://m.en.wikipedia.org/wiki/Pluto"],
        )

        assert documents.count() == 1

        actual = documents.first()
        assert isinstance(actual, document.Document)
        assert len(actual.document_uris) == 1

        docuri = actual.document_uris[0]
        assert docuri.claimant == "https://en.wikipedia.org/wiki/Pluto"
        assert docuri.uri == "https://en.wikipedia.org/wiki/Pluto"
        assert docuri.type == "self-claim"

    def test_raises_retryable_error_when_flush_fails(self, db_session, monkeypatch):
        def err():
            raise sa.exc.IntegrityError(None, None, None)

        monkeypatch.setattr(db_session, "flush", err)

        with pytest.raises(document.ConcurrentUpdateError):
            with db_session.no_autoflush:  # prevent premature IntegrityError
                document.Document.find_or_create_by_uris(
                    db_session,
                    "https://en.wikipedia.org/wiki/Pluto",
                    ["https://m.en.wikipedia.org/wiki/Pluto"],
                )


class TestDocumentWebURI(object):
    """Unit tests for Document.web_uri and Document.update_web_uri()."""

    def test_web_uri_is_initially_None(self, factories):
        assert factories.Document().web_uri is None

    @pytest.mark.parametrize(
        "document_uris,expected_web_uri",
        [
            # Given a single http or https URL it just uses it.
            ([("http://example.com", "self-claim")], "http://example.com"),
            ([("https://example.com", "self-claim")], "https://example.com"),
            ([("http://example.com", "rel-canonical")], "http://example.com"),
            ([("https://example.com", "rel-canonical")], "https://example.com"),
            ([("http://example.com", "rel-shortlink")], "http://example.com"),
            ([("https://example.com", "rel-shortlink")], "https://example.com"),
            # Given no http or https URLs it sets web_uri to None.
            ([], None),
            (
                [
                    ("ftp://example.com", "self-claim"),
                    ("android-app://example.com", "rel-canonical"),
                    ("urn:x-pdf:example", "rel-alternate"),
                    ("doi:http://example.com", "rel-shortlink"),
                ],
                None,
            ),
            # It prefers self-claim URLs over all other URLs.
            (
                [
                    ("https://example.com/shortlink", "rel-shortlink"),
                    ("https://example.com/canonical", "rel-canonical"),
                    ("https://example.com/self-claim", "self-claim"),
                ],
                "https://example.com/self-claim",
            ),
            # It prefers canonical URLs over all other non-self-claim URLs.
            (
                [
                    ("https://example.com/shortlink", "rel-shortlink"),
                    ("https://example.com/canonical", "rel-canonical"),
                ],
                "https://example.com/canonical",
            ),
            # If there's no self-claim or canonical URL it will return an https
            # URL of a different type.
            (
                [
                    ("ftp://example.com", "self-claim"),
                    ("urn:x-pdf:example", "rel-alternate"),
                    # This is the one that should be returned.
                    ("https://example.com/alternate", "rel-alternate"),
                    ("android-app://example.com", "rel-canonical"),
                    ("doi:http://example.com", "rel-shortlink"),
                ],
                "https://example.com/alternate",
            ),
            # If there's no self-claim or canonical URL it will return an http
            # URL of a different type.
            (
                [
                    ("ftp://example.com", "self-claim"),
                    ("urn:x-pdf:example", "rel-alternate"),
                    # This is the one that should be returned.
                    ("http://example.com/alternate", "rel-alternate"),
                    ("android-app://example.com", "rel-canonical"),
                    ("doi:http://example.com", "rel-shortlink"),
                ],
                "http://example.com/alternate",
            ),
        ],
    )
    def test_update_web_uri(self, document_uris, factories, expected_web_uri):
        document = factories.Document()

        for docuri_tuple in document_uris:
            factories.DocumentURI(
                uri=docuri_tuple[0], type=docuri_tuple[1], document=document
            )

        document.update_web_uri()

        assert document.web_uri == expected_web_uri


class TestDocumentURI(object):
    def test_type_defaults_to_empty_string(self, db_session):
        document_uri = document.DocumentURI(
            claimant="http://www.example.com",
            uri="http://www.example.com",
            type=None,
            content_type="bar",
            document=document.Document(),
        )
        db_session.add(document_uri)

        db_session.flush()

        assert document_uri.type == ""

    def test_you_cannot_set_type_to_null(self, db_session):
        document_uri = document.DocumentURI(
            claimant="http://www.example.com",
            uri="http://www.example.com",
            type="foo",
            content_type="bar",
            document=document.Document(),
        )
        db_session.add(document_uri)
        db_session.flush()

        document_uri.type = None

        with pytest.raises(sa.exc.IntegrityError):
            db_session.flush()

    def test_content_type_defaults_to_empty_string(self, db_session):
        document_uri = document.DocumentURI(
            claimant="http://www.example.com",
            uri="http://www.example.com",
            type="bar",
            content_type=None,
            document=document.Document(),
        )
        db_session.add(document_uri)

        db_session.flush()

        assert document_uri.content_type == ""

    def test_you_cannot_set_content_type_to_null(self, db_session):
        document_uri = document.DocumentURI(
            claimant="http://www.example.com",
            uri="http://www.example.com",
            type="foo",
            content_type="bar",
            document=document.Document(),
        )
        db_session.add(document_uri)
        db_session.flush()

        document_uri.content_type = None

        with pytest.raises(sa.exc.IntegrityError):
            db_session.flush()

    def test_you_cannot_add_duplicate_document_uris(self, db_session):
        """
        You can't add duplicate DocumentURI's to the database.

        You can't add DocumentURI's with the same claimant, uri, type and
        content_type, even if they have different documents.

        """
        db_session.add_all(
            [
                document.DocumentURI(
                    claimant="http://www.example.com",
                    uri="http://www.example.com",
                    type="foo",
                    content_type="bar",
                    document=document.Document(),
                ),
                document.DocumentURI(
                    claimant="http://www.example.com",
                    uri="http://www.example.com",
                    type="foo",
                    content_type="bar",
                    document=document.Document(),
                ),
            ]
        )

        with pytest.raises(sa.exc.IntegrityError):
            db_session.commit()


@pytest.mark.usefixtures("log")
class TestCreateOrUpdateDocumentURI(object):
    def test_it_updates_the_existing_DocumentURI_if_there_is_one(self, db_session):
        claimant = "http://example.com/example_claimant.html"
        uri = "http://example.com/example_uri.html"
        type_ = "self-claim"
        content_type = ""
        document_ = document.Document()
        created = yesterday()
        updated = yesterday()
        document_uri = document.DocumentURI(
            claimant=claimant,
            uri=uri,
            type=type_,
            content_type=content_type,
            document=document_,
            created=created,
            updated=updated,
        )
        db_session.add(document_uri)

        now_ = now()
        document.create_or_update_document_uri(
            session=db_session,
            claimant=claimant,
            uri=uri,
            type=type_,
            content_type=content_type,
            document=document_,
            created=now_,
            updated=now_,
        )

        assert document_uri.created == created
        assert document_uri.updated == now_
        assert (
            len(db_session.query(document.DocumentURI).all()) == 1
        ), "It shouldn't have added any new objects to the db"

    def test_it_creates_a_new_DocumentURI_if_there_is_no_existing_one(self, db_session):
        claimant = "http://example.com/example_claimant.html"
        uri = "http://example.com/example_uri.html"
        type_ = "self-claim"
        content_type = ""
        document_ = document.Document()
        created = yesterday()
        updated = yesterday()

        # Add one non-matching DocumentURI to the database.
        db_session.add(
            document.DocumentURI(
                claimant=claimant,
                uri=uri,
                type=type_,
                # Different content_type means this DocumentURI should not match
                # the query.
                content_type="different",
                document=document_,
                created=created,
                updated=updated,
            )
        )

        document.create_or_update_document_uri(
            session=db_session,
            claimant=claimant,
            uri=uri,
            type=type_,
            content_type=content_type,
            document=document_,
            created=now(),
            updated=now(),
        )

        document_uri = (
            db_session.query(document.DocumentURI)
            .order_by(document.DocumentURI.created.desc())
            .first()
        )
        assert document_uri.claimant == claimant
        assert document_uri.uri == uri
        assert document_uri.type == type_
        assert document_uri.content_type == content_type
        assert document_uri.document == document_
        assert document_uri.created > created
        assert document_uri.updated > updated

    def test_it_skips_denormalizing_http_s_uri_to_document(self, db_session):
        document_ = document.Document(web_uri="http://example.com/first_uri.html")
        db_session.add(document_)

        document.create_or_update_document_uri(
            session=db_session,
            claimant="http://example.com/example_claimant.html",
            uri="http://example.com/second_uri.html",
            type="self-claim",
            content_type="",
            document=document_,
            created=now(),
            updated=now(),
        )

        document_ = db_session.query(document.Document).get(document_.id)
        assert document_.web_uri == "http://example.com/first_uri.html"

    def test_it_logs_a_warning_if_document_ids_differ(self, log):
        """
        It should log a warning on Document objects mismatch.

        If there's an existing DocumentURI and its .document property is
        different to the given document it shoulg log a warning.

        """
        session = mock_db_session()

        # existing_document_uri.document won't be equal to the given document.
        existing_document_uri = mock.Mock(document=mock_document())
        session.query.return_value.filter.return_value.first.return_value = (
            existing_document_uri
        )

        document.create_or_update_document_uri(
            session=session,
            claimant="http://example.com/example_claimant.html",
            uri="http://example.com/example_uri.html",
            type="self-claim",
            content_type=None,
            document=mock_document(),
            created=now(),
            updated=now(),
        )

        assert log.warning.call_count == 1

    def test_raises_retryable_error_when_flush_fails(self, db_session, monkeypatch):
        document_ = document.Document()

        def err():
            raise sa.exc.IntegrityError(None, None, None)

        monkeypatch.setattr(db_session, "flush", err)

        with pytest.raises(document.ConcurrentUpdateError):
            with db_session.no_autoflush:  # prevent premature IntegrityError
                document.create_or_update_document_uri(
                    session=db_session,
                    claimant="http://example.com",
                    uri="http://example.org",
                    type="rel-canonical",
                    content_type="text/html",
                    document=document_,
                    created=now(),
                    updated=now(),
                )


class TestCreateOrUpdateDocumentMeta(object):
    def test_it_creates_a_new_DocumentMeta_if_there_is_no_existing_one(
        self, db_session
    ):
        claimant = "http://example.com/claimant"
        type_ = "title"
        value = "the title"
        document_ = document.Document()
        created = yesterday()
        updated = now()

        # Add one non-matching DocumentMeta to the database.
        # This should be ignored.
        db_session.add(
            document.DocumentMeta(
                claimant=claimant,
                # Different type means this should not match the query.
                type="different",
                value=value,
                document=document_,
                created=created,
                updated=updated,
            )
        )

        document.create_or_update_document_meta(
            session=db_session,
            claimant=claimant,
            type=type_,
            value=value,
            document=document_,
            created=created,
            updated=updated,
        )

        document_meta = db_session.query(document.DocumentMeta).all()[-1]
        assert document_meta.claimant == claimant
        assert document_meta.type == type_
        assert document_meta.value == value
        assert document_meta.document == document_
        assert document_meta.created == created
        assert document_meta.updated == updated

    def test_it_updates_an_existing_DocumentMeta_if_there_is_one(self, db_session):
        claimant = "http://example.com/claimant"
        type_ = "title"
        value = "the title"
        document_ = document.Document()
        created = yesterday()
        updated = now()
        document_meta = document.DocumentMeta(
            claimant=claimant,
            type=type_,
            value=value,
            document=document_,
            created=created,
            updated=updated,
        )
        db_session.add(document_meta)

        new_updated = now()
        document.create_or_update_document_meta(
            session=db_session,
            claimant=claimant,
            type=type_,
            value="new value",
            document=document.Document(),  # This should be ignored.
            created=now(),  # This should be ignored.
            updated=new_updated,
        )

        assert document_meta.value == "new value"
        assert document_meta.updated == new_updated
        assert document_meta.created == created, "It shouldn't update created"
        assert document_meta.document == document_, "It shouldn't update document"
        assert (
            len(db_session.query(document.DocumentMeta).all()) == 1
        ), "It shouldn't have added any new objects to the db"

    def test_it_denormalizes_title_to_document_when_none(self, db_session):
        claimant = "http://example.com/claimant"
        type_ = "title"
        value = ["the title"]
        document_ = document.Document(title=None)
        created = yesterday()
        updated = now()
        db_session.add(document_)

        document.create_or_update_document_meta(
            session=db_session,
            claimant=claimant,
            type=type_,
            value=value,
            document=document_,
            created=created,
            updated=updated,
        )

        document_ = db_session.query(document.Document).get(document_.id)
        assert document_.title == value[0]

    def test_it_denormalizes_title_to_document_when_empty(self, db_session):
        claimant = "http://example.com/claimant"
        type_ = "title"
        value = ["the title"]
        document_ = document.Document(title="")
        created = yesterday()
        updated = now()
        db_session.add(document_)

        document.create_or_update_document_meta(
            session=db_session,
            claimant=claimant,
            type=type_,
            value=value,
            document=document_,
            created=created,
            updated=updated,
        )

        document_ = db_session.query(document.Document).get(document_.id)
        assert document_.title == value[0]

    def test_it_skips_denormalizing_title_to_document_when_already_set(
        self, db_session
    ):
        claimant = "http://example.com/claimant"
        type_ = "title"
        value = ["the title"]
        document_ = document.Document(title="foobar")
        created = yesterday()
        updated = now()
        db_session.add(document_)

        document.create_or_update_document_meta(
            session=db_session,
            claimant=claimant,
            type=type_,
            value=value,
            document=document_,
            created=created,
            updated=updated,
        )

        document_ = db_session.query(document.Document).get(document_.id)
        assert document_.title == "foobar"

    def test_it_logs_a_warning(self, log):
        """
        It should warn on document mismatches.

        It should warn if there's an existing DocumentMeta with a different
        Document.

        """
        session = mock_db_session()
        document_one = mock_document()
        document_two = mock_document()
        existing_document_meta = mock_document_meta(document=document_one)
        session.query.return_value.filter.return_value.one_or_none.return_value = (
            existing_document_meta
        )

        document.create_or_update_document_meta(
            session=session,
            claimant="http://example.com/claimant",
            type="title",
            value="new value",
            document=document_two,
            created=yesterday(),
            updated=now(),
        )

        assert log.warning.call_count == 1

    def test_raises_retryable_error_when_flush_fails(self, db_session, monkeypatch):
        document_ = document.Document()

        def err():
            raise sa.exc.IntegrityError(None, None, None)

        monkeypatch.setattr(db_session, "flush", err)

        with pytest.raises(document.ConcurrentUpdateError):
            with db_session.no_autoflush:  # prevent premature IntegrityError
                document.create_or_update_document_meta(
                    session=db_session,
                    claimant="http://example.com",
                    type="title",
                    value="My Title",
                    document=document_,
                    created=now(),
                    updated=now(),
                )


@pytest.mark.usefixtures("merge_data")
class TestMergeDocuments(object):
    def test_merge_documents_returns_master(self, db_session, merge_data):
        master, _, _ = merge_data

        merged_master = document.merge_documents(db_session, merge_data)

        assert merged_master == master

    def test_merge_documents_deletes_duplicate_documents(self, db_session, merge_data):
        _, duplicate_1, duplicate_2 = merge_data

        document.merge_documents(db_session, merge_data)
        db_session.flush()

        count = (
            db_session.query(document.Document)
            .filter(document.Document.id.in_([duplicate_1.id, duplicate_2.id]))
            .count()
        )

        assert count == 0

    def test_merge_documents_rewires_document_uris(self, db_session, merge_data):
        master, duplicate_1, duplicate_2 = merge_data

        document.merge_documents(db_session, merge_data)
        db_session.flush()

        assert len(master.document_uris) == 3
        assert len(duplicate_1.document_uris) == 0
        assert len(duplicate_2.document_uris) == 0

    def test_merge_documents_rewires_document_meta(self, db_session, merge_data):
        master, duplicate_1, duplicate_2 = merge_data

        document.merge_documents(db_session, merge_data)
        db_session.flush()

        assert len(master.meta) == 3
        assert len(duplicate_1.meta) == 0
        assert len(duplicate_2.meta) == 0

    def test_merge_documents_rewires_annotations(self, db_session, merge_data):
        master, duplicate_1, duplicate_2 = merge_data

        document.merge_documents(db_session, merge_data)
        db_session.flush()

        assert (
            6
            == db_session.query(models.Annotation)
            .filter_by(document_id=master.id)
            .count()
        )
        assert (
            0
            == db_session.query(models.Annotation)
            .filter_by(document_id=duplicate_1.id)
            .count()
        )
        assert (
            0
            == db_session.query(models.Annotation)
            .filter_by(document_id=duplicate_2.id)
            .count()
        )

    def test_raises_retryable_error_when_flush_fails(
        self, db_session, merge_data, monkeypatch
    ):
        def err():
            raise sa.exc.IntegrityError(None, None, None)

        monkeypatch.setattr(db_session, "flush", err)

        with pytest.raises(document.ConcurrentUpdateError):
            document.merge_documents(db_session, merge_data)

    @pytest.fixture
    def merge_data(self, db_session, request):
        master = document.Document(
            document_uris=[
                document.DocumentURI(
                    claimant="https://en.wikipedia.org/wiki/Main_Page",
                    uri="https://en.wikipedia.org/wiki/Main_Page",
                    type="self-claim",
                )
            ],
            meta=[
                document.DocumentMeta(
                    claimant="https://en.wikipedia.org/wiki/Main_Page",
                    type="title",
                    value="Wikipedia, the free encyclopedia",
                )
            ],
        )
        duplicate_1 = document.Document(
            document_uris=[
                document.DocumentURI(
                    claimant="https://m.en.wikipedia.org/wiki/Main_Page",
                    uri="https://en.wikipedia.org/wiki/Main_Page",
                    type="rel-canonical",
                )
            ],
            meta=[
                document.DocumentMeta(
                    claimant="https://m.en.wikipedia.org/wiki/Main_Page",
                    type="title",
                    value="Wikipedia, the free encyclopedia",
                )
            ],
        )
        duplicate_2 = document.Document(
            document_uris=[
                document.DocumentURI(
                    claimant="https://en.wikipedia.org/wiki/Home",
                    uri="https://en.wikipedia.org/wiki/Main_Page",
                    type="rel-canonical",
                )
            ],
            meta=[
                document.DocumentMeta(
                    claimant="https://en.wikipedia.org/wiki/Home",
                    type="title",
                    value="Wikipedia, the free encyclopedia",
                )
            ],
        )

        db_session.add_all([master, duplicate_1, duplicate_2])
        db_session.flush()

        master_ann_1 = models.Annotation(userid="luke", document_id=master.id)
        master_ann_2 = models.Annotation(userid="alice", document_id=master.id)
        duplicate_1_ann_1 = models.Annotation(userid="lucy", document_id=duplicate_1.id)
        duplicate_1_ann_2 = models.Annotation(userid="bob", document_id=duplicate_1.id)
        duplicate_2_ann_1 = models.Annotation(userid="amy", document_id=duplicate_2.id)
        duplicate_2_ann_2 = models.Annotation(userid="dan", document_id=duplicate_2.id)
        db_session.add_all(
            [
                master_ann_1,
                master_ann_2,
                duplicate_1_ann_1,
                duplicate_1_ann_2,
                duplicate_2_ann_1,
                duplicate_2_ann_2,
            ]
        )
        return (master, duplicate_1, duplicate_2)


class TestUpdateDocumentMetadata(object):
    def test_it_uses_the_target_uri_to_get_the_document(
        self, annotation, Document, session
    ):
        document_uri_dicts = [
            {
                "uri": "http://example.com/example_1",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
            {
                "uri": "http://example.com/example_2",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
            {
                "uri": "http://example.com/example_3",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
        ]

        document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            document_uri_dicts,
            annotation.created,
            annotation.updated,
        )

        Document.find_or_create_by_uris.assert_called_once_with(
            session,
            annotation.target_uri,
            [
                "http://example.com/example_1",
                "http://example.com/example_2",
                "http://example.com/example_3",
            ],
            created=annotation.created,
            updated=annotation.updated,
        )

    def test_if_there_are_multiple_documents_it_merges_them_into_one(
        self, annotation, Document, merge_documents, session
    ):
        """If it finds more than one document it calls merge_documents()."""
        Document.find_or_create_by_uris.return_value = mock.Mock(
            count=mock.Mock(return_value=3)
        )

        document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            [],
            annotation.created,
            annotation.updated,
        )

        merge_documents.assert_called_once_with(
            session,
            Document.find_or_create_by_uris.return_value,
            updated=annotation.updated,
        )

    def test_it_calls_first(self, annotation, session, Document):
        """If it finds only one document it calls first()."""
        Document.find_or_create_by_uris.return_value = mock.Mock(
            count=mock.Mock(return_value=1)
        )

        document.update_document_metadata(session, annotation, [], [])

        Document.find_or_create_by_uris.return_value.first.assert_called_once_with()

    def test_it_updates_document_updated(
        self, annotation, Document, merge_documents, session
    ):
        yesterday_ = "yesterday"
        document_ = merge_documents.return_value = mock.Mock(updated=yesterday_)
        Document.find_or_create_by_uris.return_value.first.return_value = document_

        document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            [],
            annotation.created,
            annotation.updated,
        )

        assert document_.updated == annotation.updated

    def test_it_saves_all_the_document_uris(
        self, session, annotation, Document, create_or_update_document_uri
    ):
        """It creates or updates a DocumentURI for each document URI dict."""
        Document.find_or_create_by_uris.return_value.count.return_value = 1

        document_uri_dicts = [
            {
                "uri": "http://example.com/example_1",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
            {
                "uri": "http://example.com/example_2",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
            {
                "uri": "http://example.com/example_3",
                "claimant": "http://example.com/claimant",
                "type": "type",
                "content_type": None,
            },
        ]

        document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            document_uri_dicts,
            annotation.created,
            annotation.updated,
        )

        assert create_or_update_document_uri.call_count == 3
        for doc_uri_dict in document_uri_dicts:
            create_or_update_document_uri.assert_any_call(
                session=session,
                document=Document.find_or_create_by_uris.return_value.first.return_value,
                created=annotation.created,
                updated=annotation.updated,
                **doc_uri_dict
            )

    def test_it_updates_document_web_uri(
        self, annotation, Document, factories, session
    ):
        document_ = mock.Mock(web_uri=None)
        Document.find_or_create_by_uris.return_value.count.return_value = 1
        Document.find_or_create_by_uris.return_value.first.return_value = document_

        document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            [],
            annotation.created,
            annotation.updated,
        )

        document_.update_web_uri.assert_called_once_with()

    def test_it_saves_all_the_document_metas(
        self, annotation, create_or_update_document_meta, Document, session
    ):
        """It creates or updates a DocumentMeta for each document meta dict."""
        Document.find_or_create_by_uris.return_value.count.return_value = 1

        document_meta_dicts = [
            {
                "claimant": "http://example.com/claimant",
                "type": "title",
                "value": "foo",
            },
            {
                "type": "article title",
                "value": "bar",
                "claimant": "http://example.com/claimant",
            },
            {
                "type": "site title",
                "value": "gar",
                "claimant": "http://example.com/claimant",
            },
        ]

        document.update_document_metadata(
            session,
            annotation.target_uri,
            document_meta_dicts,
            [],
            annotation.created,
            annotation.updated,
        )

        assert create_or_update_document_meta.call_count == 3
        for document_meta_dict in document_meta_dicts:
            create_or_update_document_meta.assert_any_call(
                session=session,
                document=Document.find_or_create_by_uris.return_value.first.return_value,
                created=annotation.created,
                updated=annotation.updated,
                **document_meta_dict
            )

    def test_it_returns_a_document(
        self, annotation, create_or_update_document_meta, Document, session
    ):
        Document.find_or_create_by_uris.return_value.count.return_value = 1

        result = document.update_document_metadata(
            session,
            annotation.target_uri,
            [],
            [],
            annotation.created,
            annotation.updated,
        )

        assert result == Document.find_or_create_by_uris.return_value.first.return_value

    @pytest.fixture
    def annotation(self):
        return mock.Mock(spec=models.Annotation())

    @pytest.fixture
    def create_or_update_document_meta(self, patch):
        return patch("hyputils.memex.models.document.create_or_update_document_meta")

    @pytest.fixture
    def create_or_update_document_uri(self, patch):
        return patch("hyputils.memex.models.document.create_or_update_document_uri")

    @pytest.fixture
    def Document(self, patch):
        Document = patch("hyputils.memex.models.document.Document")
        Document.find_or_create_by_uris.return_value.count.return_value = 1
        return Document

    @pytest.fixture
    def merge_documents(self, patch):
        return patch("hyputils.memex.models.document.merge_documents")

    @pytest.fixture
    def session(self, db_session):
        return mock.Mock(spec=db_session)


def now():
    return datetime.datetime.now()


def yesterday():
    return now() - datetime.timedelta(days=1)


def mock_db_session():
    """Return a mock db session object."""

    class DB(object):
        def add(self, obj):
            pass

        def query(self, cls):
            pass

        def flush(self):
            pass

    return mock.Mock(spec=DB())


def mock_document():
    """Return a mock Document object."""
    return mock.Mock(spec=document.Document())


def mock_document_meta(document=None):

    # We define a class to use as the mock spec here because we can't use the
    # real DocumentMeta class because that class may be patched in the tests
    # that are calling this function (so we'd end up using a mock object as a
    # spec instead, and get completely the wrong spec).
    class DocumentMeta(object):
        def __init__(self):
            self.type = None
            self.value = None
            self.created = None
            self.updated = None
            self.document = document
            self.id = None
            self.document_id = None

    return mock.Mock(spec=DocumentMeta())


@pytest.fixture
def log(patch):
    return patch("hyputils.memex.models.document.log")



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import factory

from hyputils.memex import models
from hyputils.memex.models.group import JoinableBy, ReadableBy, WriteableBy

from .base import ModelFactory
from .user import User


class Group(ModelFactory):
    class Meta:
        model = models.Group
        sqlalchemy_session_persistence = "flush"

    name = factory.Sequence(lambda n: "Test Group {n}".format(n=str(n)))
    authority = "example.com"
    creator = factory.SubFactory(User)
    joinable_by = JoinableBy.authority
    readable_by = ReadableBy.members
    writeable_by = WriteableBy.members
    members = factory.LazyAttribute(lambda obj: [obj.creator])


class OpenGroup(Group):

    name = factory.Sequence(lambda n: "Test Open Group {n}".format(n=str(n)))

    joinable_by = None
    readable_by = ReadableBy.world
    writeable_by = WriteableBy.authority
    members = []


class RestrictedGroup(Group):
    name = factory.Sequence(lambda n: "Test Restricted Group {n}".format(n=str(n)))

    joinable_by = None
    readable_by = ReadableBy.world



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import factory
import faker

FAKER = faker.Factory.create()
SESSION = None


def set_session(value):
    global SESSION

    SESSION = value


class ModelFactory(factory.alchemy.SQLAlchemyModelFactory):
    class Meta:
        abstract = True

    @classmethod
    def _create(cls, model_class, *args, **kwargs):
        # We override SQLAlchemyModelFactory's default _create classmethod so
        # that rather than fetching the session from cls._meta (which is
        # created at parse time... ugh) we fetch it from the SESSION global,
        # which is dynamically filled out by the `factories` fixture when
        # used.
        if SESSION is None:
            raise RuntimeError("no session: did you use the factories fixture?")
        obj = model_class(*args, **kwargs)
        SESSION.add(obj)
        if cls._meta.sqlalchemy_session_persistence == "flush":
            SESSION.flush()
        return obj



# -*- coding: utf-8 -*-

"""Factory classes for easily generating test objects."""
from __future__ import unicode_literals

from .base import set_session

from .annotation import Annotation
from .document import Document, DocumentMeta, DocumentURI
from .group import Group, OpenGroup, RestrictedGroup
from .organization import Organization
from .user import User
from .user_identity import UserIdentity

__all__ = (
    "Annotation",
    "Document",
    "DocumentMeta",
    "DocumentURI",
    "Group",
    "OpenGroup",
    "Organization",
    "RestrictedGroup",
    "User",
    "UserIdentity",
    "set_session",
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import pytest
import sqlalchemy.exc

from hyputils.memex import models
from hyputils.memex._compat import PY2


class TestUserIdentity(object):
    def test_you_can_save_and_then_retrieve_field_values(
        self, db_session, matchers, user
    ):
        user_identity_1 = models.UserIdentity(
            provider="provider_1", provider_unique_id="1", user=user
        )
        user_identity_2 = models.UserIdentity(
            provider="provider_1", provider_unique_id="2", user=user
        )
        user_identity_3 = models.UserIdentity(
            provider="provider_2", provider_unique_id="3", user=user
        )

        db_session.add_all([user_identity_1, user_identity_2, user_identity_3])
        db_session.flush()

        user_identities = (
            db_session.query(models.UserIdentity)
            .order_by(models.UserIdentity.provider_unique_id)
            .all()
        )

        # Auto incrementing unique IDs should have been generated for us.
        assert type(user_identities[0].id) is int
        assert type(user_identities[1].id) is int
        assert type(user_identities[2].id) is int

        # The provider strings that we gave should have been saved.
        assert user_identities[0].provider == "provider_1"
        assert user_identities[1].provider == "provider_1"
        assert user_identities[2].provider == "provider_2"

        # The provider_unique_id strings that we gave should have been saved.
        assert user_identities[0].provider_unique_id == "1"
        assert user_identities[1].provider_unique_id == "2"
        assert user_identities[2].provider_unique_id == "3"

    def test_provider_cant_be_null(self, db_session, user):
        db_session.add(models.UserIdentity(provider_unique_id="1", user=user))

        with pytest.raises(
            sqlalchemy.exc.IntegrityError,
            match='null value in column "provider" violates not-null constraint',
        ):
            db_session.flush()

    def test_provider_id_cant_be_null(self, db_session, user):
        db_session.add(models.UserIdentity(provider="provider", user=user))

        with pytest.raises(
            sqlalchemy.exc.IntegrityError,
            match='null value in column "provider_unique_id" violates not-null constraint',
        ):
            db_session.flush()

    def test_user_cant_be_null(self, db_session):
        db_session.add(models.UserIdentity(provider="provider", provider_unique_id="1"))

        with pytest.raises(
            sqlalchemy.exc.IntegrityError,
            match='null value in column "user_id" violates not-null constraint',
        ):
            db_session.flush()

    def test_two_cant_have_the_same_provider_and_provider_id(
        self, db_session, factories
    ):
        db_session.add_all(
            [
                models.UserIdentity(
                    provider="provider", provider_unique_id="id", user=factories.User()
                ),
                models.UserIdentity(
                    provider="provider", provider_unique_id="id", user=factories.User()
                ),
            ]
        )

        with pytest.raises(
            sqlalchemy.exc.IntegrityError,
            match='duplicate key value violates unique constraint "uq__user_identity__provider"',
        ):
            db_session.flush()

    def test_one_user_can_have_the_same_provider_id_from_different_providers(
        self, db_session, user
    ):
        db_session.add_all(
            [
                models.UserIdentity(
                    provider="provider_1", provider_unique_id="id", user=user
                ),
                models.UserIdentity(
                    provider="provider_2", provider_unique_id="id", user=user
                ),
            ]
        )

        db_session.flush()

    def test_different_users_can_have_the_same_provider_id_from_different_providers(
        self, db_session, factories
    ):
        db_session.add_all(
            [
                models.UserIdentity(
                    provider="provider_1",
                    provider_unique_id="id",
                    user=factories.User(),
                ),
                models.UserIdentity(
                    provider="provider_2",
                    provider_unique_id="id",
                    user=factories.User(),
                ),
            ]
        )

        db_session.flush()

    def test_removing_a_user_identity_from_a_user_deletes_the_user_identity_from_the_db(
        self, db_session, user
    ):
        # Add a couple of noise UserIdentity's. These should not be removed
        # from the DB.
        models.UserIdentity(provider="provider", provider_unique_id="1", user=user)
        models.UserIdentity(provider="provider", provider_unique_id="2", user=user)
        # The UserIdentity that we are going to remove.
        user_identity = models.UserIdentity(
            provider="provider", provider_unique_id="3", user=user
        )

        user.identities.remove(user_identity)

        assert user_identity not in db_session.query(models.UserIdentity).all()

    def test_deleting_a_user_identity_removes_it_from_its_user(self, db_session, user):
        # Add a couple of noise UserIdentity's. These should not be removed
        # from user.identities.
        models.UserIdentity(provider="provider", provider_unique_id="1", user=user)
        models.UserIdentity(provider="provider", provider_unique_id="2", user=user)
        # The UserIdentity that we are going to remove.
        user_identity = models.UserIdentity(
            provider="provider", provider_unique_id="3", user=user
        )
        db_session.commit()

        db_session.delete(user_identity)

        db_session.refresh(user)  # Make sure user.identities is up to date.
        assert user_identity not in user.identities

    def test_deleting_a_user_deletes_all_its_user_identities(self, db_session, user):
        models.UserIdentity(provider="provider", provider_unique_id="1", user=user)
        models.UserIdentity(provider="provider", provider_unique_id="2", user=user)
        db_session.commit()

        db_session.delete(user)

        assert db_session.query(models.UserIdentity).count() == 0

    def test_repr(self):
        user_identity = models.UserIdentity(
            provider="provider_1", provider_unique_id="1"
        )

        expected_repr = "UserIdentity(provider='provider_1', provider_unique_id='1')"

        if PY2:
            expected_repr = (
                "UserIdentity(provider=u'provider_1', " "provider_unique_id=u'1')"
            )

        assert repr(user_identity) == expected_repr

    @pytest.fixture
    def user(self, factories):
        return factories.User()



"""
    Implementation of various filter/handler pairs.
"""

import os


class filterHandler:
    """ Base class that all filter handlers should be derived from.
        Any authenticiation state needed to link services should be
        managed in __init__. Sorta a crappy python ITTT.
    """

    def filter(self, message):
        return True

    def handler(self, message):
        raise NotImplemented('You need to implement this in a subclass')

    def __call__(self, message):
        if self.filter(message):
            self.handler(message)


class dbSyncHandler(filterHandler):
    def __init__(self, *helpers):
        self.helpers = helpers

    def handler(self, message):
        for helper in self.helpers:
            helper(message)


class annotationSyncHandler(filterHandler):
    class DeletedAnno:
        deleted = True
        def __init__(self, id):
            self.id = id

    def __init__(self, annos, memoizer=None):
        from .hypothesis import HypothesisAnnotation as ha
        self.HypothesisAnnotation = ha
        self.annos = annos
        if memoizer is not None:
            self.memoizer = memoizer
        if not hasattr(self, 'memoizer'):
            print(f'WARNING: no memoizer has been supplied for {self.__class__.__name__}')

    def handler(self, message):
        anno = None
        try:
            act = message['options']['action']
            if act != 'create': # update delete
                mid = message['payload'][0]['id']
                for gone in [_ for _ in self.annos if _.id == mid]:
                    # FIXME slow, replace when we switch over to using dicts
                    self.annos.remove(gone)
                anno = self.DeletedAnno(mid)
            if act != 'delete':  # create update
                anno = self.HypothesisAnnotation(message['payload'][0])
                self.annos.append(anno)
            #print(len(self.annos), 'annotations.')
            if hasattr(self, 'memoizer'):
                self.memoizer.memoize_annos(self.annos)

            return anno  # we can't not return None
        except KeyError as e:
            embed()


class helperSyncHandler(annotationSyncHandler):
    def __init__(self, annos, *helpers, memoizer=None):
        super().__init__(annos, memoizer=memoizer)
        if helpers:
            self.helpers = helpers

    def handler(self, message):
        anno = super().handler(message)
        out = []  # can't use yield here

        if anno is None:
            raise TypeError(f'anno should not be None! Bad message:\n{message}')

        for helper in self.helpers:
            out.append(helper(anno, self.annos))

        return tuple(_ for _ in out if _ is not None)


class websocketServerHandler(filterHandler):
    def __init__(self, send_to_server):
        self.send = send_to_server

    def handler(self, message):
        self.send(message)


class slackHandler(filterHandler):
    def __init__(self, api_token):
        self.api_token = api_token
    
    def filter(self):
        pass

    def handler(self):
        pass


class zendeskHandler(filterHandler):
    """ Handler that creates zendeks tickets. Currently hardcoding filter
        rules (eg uri) but could make this configurabe if desired. May also
        break these filterHandlers out into their own file that users could
        create independently of the subscription setup here....
    """
    def __init__(self, infopath='~/files/zendeskinfo.yaml'):
        import yaml
        from zdesk import Zendesk
        with open(os.path.expanduser(infopath), 'rt') as f:
            self.zendeskinfo = yaml.load(f)
        self.zendesk = Zendesk(**self.zendeskinfo)

    def filter(self, message):
        if message['options']['action'] != 'create':
            print('Annotation was not a creation, run update path?')
            return
        data = message['payload'][0]
        if 'scicrunch.org' in data['uri']:
            return True

    def handler(self, message):
        data = message['payload'][0]
        exact = [i['exact'] for o in data['target'] for i in o['selector'] if 'selector' in o and 'exact' in i]
        exact = exact[0] if exact else 'No description provided'
        href = data['links']['incontext']
        new_ticket = {                                 
            'ticket': {
                'requester': {
                    'name': 'hypothesis bot',
                    'email': self.zendeskinfo['zdesk_email'],
                },
                'subject':'hypothes.is annotation on {} by {}'.format(data['document']['title'], data['user'].split(':',1)[1]),
                'description': '"{}" | {} {}'.format(exact, data['text'], href) , 
                'tags': data['tags'],
                #'ticket_field_entries': [
                    #{
                        #'ticket_field_id':24394906,  # referrer
                        #'value': 'hypothes.is {}'.format(data['user'].split(':',1)[1])  # hypothes.is
                    #},
                    #{
                        #'ticket_field_id':24394926,  # url
                        #'ticket_field_id':'Url',  # url
                        #'value': data['links']['incontext']  # incontext_url
                    #}
                #]
            }
        }
        result = self.zendesk.ticket_create(data=new_ticket)
        print(result)


class printHandler(filterHandler):
    def filter(self, message):
        return True

    def handler(self, message):
        #print(message['payload'][0]['links']['incontext'])  # structure changed?
        print(message)


def getFilterHandlers():
    return [zendeskHandler(), printHandler()]  # order matters




# -*- coding: utf-8 -*-
"""Some shared utility functions for manipulating user data."""
from __future__ import unicode_literals
import re


def split_user(userid):
    """Return the user and domain parts from the given user id as a dict.

    For example if userid is u'acct:seanh@hypothes.is' then return
    {'username': u'seanh', 'domain': u'hypothes.is'}'

    :raises ValueError: if the given userid isn't a valid userid

    """
    match = re.match(r"^acct:([^@]+)@(.*)$", userid)
    if match:
        return {"username": match.groups()[0], "domain": match.groups()[1]}
    raise ValueError("{userid} isn't a valid userid".format(userid=userid))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import pytest

from sqlalchemy.dialects.postgresql import dialect

from hyputils.memex.db import types

UUID_FIXTURES = [
    # 16 byte UUIDv4s
    ("OHFscuRvTmiwpOHWfrbzxw", "38716c72e46f4e68b0a4e1d67eb6f3c7"),
    ("JCfVA8bARx6wdjQq-vp9VA", "2427d503c6c0471eb076342afafa7d54"),
    ("3AlOaWxzQBKYjRtK3_Uyvw", "dc094e696c734012988d1b4adff532bf"),
    # 15-byte ElasticSearch flake IDs
    ("AVGwR1YD8sFu_DXLVRKw", "0151b0475603ef2c516efc35cb5512b0"),
    ("AVGy6_TM8sFu_DXLVRfp", "0151b2ebf4ccef2c516efc35cb5517e9"),
    ("AVGM9RGq8sFu_DXLVN2l", "01518cf511aaef2c516efc35cb54dda5"),
    # Null values
    (None, None),
]

UUID_FIXTURES_INVALID = [
    # Incorrect type
    123,
    [1, 2, 3],
    # 20 bytes of not-base64-encoded data
    "!!!@@@^abcdef&@@@!!!",
    # 22 bytes of not-base64-encoded data
    "!!!@@@^^abcdef&&@@@!!!",
    # short data
    "MDEyMzQ1Njc4OTAxMjM=",
    "MDEyMzQ1Njc4OTAxMjM0==",
    # long data
    "MDEyMzQ1Njc4OTAxMjM0NTY=",
    "YWxsIGtpbmRzIG9mIHdlaXJkIHN0dWZmIGdvaW5nIG9u",
]


@pytest.mark.parametrize("app,db", UUID_FIXTURES)
def test_uuid_serialize(app, db):
    t = types.URLSafeUUID()
    assert t.process_bind_param(app, dialect) == db


@pytest.mark.parametrize("data", UUID_FIXTURES_INVALID)
def test_uuid_serialize_non_base64_data(data):
    t = types.URLSafeUUID()
    with pytest.raises(types.InvalidUUID):
        t.process_bind_param(data, dialect)


@pytest.mark.parametrize("app,db", UUID_FIXTURES)
def test_uuid_deserialize(app, db):
    t = types.URLSafeUUID()
    assert t.process_result_value(db, dialect) == app


def test_annotation_selector_serialize():
    t = types.AnnotationSelectorJSONB()
    selectors = [
        {
            "type": "TextQuoteSelector",
            "prefix": "\u0000Lorem ipsum ",
            "exact": "dolor sit amet,\u0000 ",
            "suffix": "consectetur\u0000 adipiscing elit.",
        }
    ]

    value = t.process_bind_param(selectors, dialect)
    assert value[0]["prefix"] == "\\u0000Lorem ipsum "
    assert value[0]["exact"] == "dolor sit amet,\\u0000 "
    assert value[0]["suffix"] == "consectetur\\u0000 adipiscing elit."


def test_annotation_selector_serialize_missing_text_quote_selector():
    t = types.AnnotationSelectorJSONB()
    selectors = [
        {
            "type": "RangeSelector",
            "startContainer": "/div[1]/div[2]",
            "endContainer": "/div[1]/div[3]",
            "startOffset": 39,
            "endoffset": 1,
        }
    ]
    assert t.process_bind_param(selectors, dialect) == selectors


def test_annotation_selector_deserialize():
    t = types.AnnotationSelectorJSONB()
    selectors = [
        {
            "type": "TextQuoteSelector",
            "prefix": "\\u0000Lorem ipsum ",
            "exact": "dolor sit amet,\\u0000 ",
            "suffix": "consectetur\\u0000 adipiscing elit.",
        }
    ]

    value = t.process_result_value(selectors, dialect)
    assert value[0]["prefix"] == "\u0000Lorem ipsum "
    assert value[0]["exact"] == "dolor sit amet,\u0000 "
    assert value[0]["suffix"] == "consectetur\u0000 adipiscing elit."


def test_annotation_selector_deserialize_missing_text_quote_selector():
    t = types.AnnotationSelectorJSONB()
    selectors = [
        {
            "type": "RangeSelector",
            "startContainer": "/div[1]/div[2]",
            "endContainer": "/div[1]/div[3]",
            "startOffset": 39,
            "endoffset": 1,
        }
    ]
    assert t.process_result_value(selectors, dialect) == selectors



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import pytest

from hyputils.memex.security import security

from hyputils.memex import models
from hyputils.memex.models.group import (
    JoinableBy,
    ReadableBy,
    WriteableBy,
    AUTHORITY_PROVIDED_ID_MAX_LENGTH,
)


def test_init_sets_given_attributes():
    group = models.Group(name="My group", authority="example.com")

    assert group.name == "My group"
    assert group.authority == "example.com"


def test_with_short_name():
    """Should raise ValueError if name shorter than 3 characters."""
    with pytest.raises(ValueError):
        models.Group(name="ab")


def test_with_long_name():
    """Should raise ValueError if name longer than 25 characters."""
    with pytest.raises(ValueError):
        models.Group(name="abcdefghijklmnopqrstuvwxyz")


def test_slug(db_session, factories, default_organization):
    name = "My Hypothesis Group"
    user = factories.User()

    group = models.Group(
        name=name,
        authority="foobar.com",
        creator=user,
        organization=default_organization,
    )
    db_session.add(group)
    db_session.flush()

    assert group.slug == "my-hypothesis-group"


def test_type_returns_open_for_open_groups(factories):
    assert factories.OpenGroup().type == "open"


def test_type_returns_private_for_private_groups(factories):
    assert factories.Group().type == "private"


def test_type_returns_restricted_for_restricted_groups(factories):
    assert factories.RestrictedGroup().type == "restricted"


def test_it_returns_None_by_default_for_authority_provided_id():
    group = models.Group(name="abcdefg")

    assert group.authority_provided_id is None


def test_it_returns_None_for_groupid_if_authority_provided_id_is_None(factories):
    group = factories.Group()

    assert group.groupid is None


def test_it_returns_formatted_groupid_if_authority_provided_id(factories):
    group = factories.Group()
    group.authority_provided_id = "hithere"

    assert group.groupid == "group:hithere@{authority}".format(
        authority=group.authority
    )


def test_groupid_setter_raises_ValueError_if_groupid_invalid(factories):
    group = factories.Group()

    with pytest.raises(ValueError, match="isn't a valid groupid"):
        group.groupid = "nonsense"


def test_groupid_setter_sets_consistuent_fields(factories):
    group = factories.Group()
    group.groupid = "group:onetwo@threefour.com"

    assert group.authority_provided_id == "onetwo"
    assert group.authority == "threefour.com"


def test_groupid_setter_accepts_None_and_nullifies_authority_provided_id(factories):
    group = factories.Group()
    group.groupid = "group:onetwo@threefour.com"
    group.groupid = None

    assert group.groupid is None
    assert group.authority == "threefour.com"
    assert group.authority_provided_id is None


@pytest.mark.parametrize(
    "authority_provided_id", ["%%&whatever", "^flop", "#---", "ßeta"]
)
def test_it_raises_ValueError_if_invalid_authority_provided_id(authority_provided_id):
    group = models.Group(name="abcdefg")

    with pytest.raises(ValueError, match="authority_provided_id must only contain"):
        group.authority_provided_id = authority_provided_id


def test_it_raises_ValueError_if_authority_provided_id_too_long():
    group = models.Group(name="abcdefg")

    with pytest.raises(ValueError, match="characters or fewer"):
        group.authority_provided_id = "a" * (AUTHORITY_PROVIDED_ID_MAX_LENGTH + 1)


def test_it_allows_authority_provided_id_to_be_None():
    group = models.Group(name="abcdefg")

    group.authority_provided_id = None


def test_type_raises_for_unknown_type_of_group(factories):
    group = factories.Group()
    # Set the group's access flags to an invalid / unused combination.
    group.joinable_by = None
    group.readable_by = ReadableBy.members
    group.writeable_by = WriteableBy.authority

    expected_err = "^This group doesn't seem to match any known type"
    with pytest.raises(ValueError, match=expected_err):
        group.type


def test_you_cannot_set_type(factories):
    group = factories.Group()

    with pytest.raises(AttributeError, match="can't set attribute"):
        group.type = "open"


def test_repr(db_session, factories, default_organization):
    name = "My Hypothesis Group"
    user = factories.User()

    group = models.Group(
        name=name,
        authority="foobar.com",
        creator=user,
        organization=default_organization,
    )
    db_session.add(group)
    db_session.flush()

    assert repr(group) == "<Group: my-hypothesis-group>"


def test_group_organization(db_session):
    name = "My Hypothesis Group"

    org = models.Organization(name="My Organization", authority="foobar.com")
    db_session.add(org)
    db_session.flush()

    group = models.Group(name=name, authority="foobar.com", organization=org)
    db_session.add(group)
    db_session.flush()

    assert group.organization == org
    assert group.organization_id == org.id


def test_created_by(db_session, factories, default_organization):
    name_1 = "My first group"
    name_2 = "My second group"
    user = factories.User()

    group_1 = models.Group(
        name=name_1,
        authority="foobar.com",
        creator=user,
        organization=default_organization,
    )
    group_2 = models.Group(
        name=name_2,
        authority="foobar.com",
        creator=user,
        organization=default_organization,
    )

    db_session.add_all([group_1, group_2])
    db_session.flush()

    assert models.Group.created_by(db_session, user).all() == [group_1, group_2]


def test_public_group():
    group = models.Group(readable_by=ReadableBy.world)

    assert group.is_public


def test_non_public_group():
    group = models.Group(readable_by=ReadableBy.members)

    assert not group.is_public


class TestGroupACL(object):
    def test_auth_client_with_matching_authority_may_add_members(
        self, group, authz_policy
    ):
        group.authority = "weewhack.com"

        assert authz_policy.permits(
            group, ["flip", "client_authority:weewhack.com"], "member_add"
        )

    def test_auth_client_without_matching_authority_may_not_add_members(
        self, group, authz_policy
    ):
        group.authority = "weewhack.com"

        assert not authz_policy.permits(
            group, ["flip", "client_authority:2weewhack.com"], "member_add"
        )

    def test_user_with_authority_may_not_add_members(self, group, authz_policy):
        group.authority = "fabuloso.biz"

        assert not authz_policy.permits(
            group, ["flip", "authority:fabuloso.biz"], "member_add"
        )

    def test_authority_joinable(self, group, authz_policy):
        group.joinable_by = JoinableBy.authority

        assert authz_policy.permits(group, ["userid", "authority:example.com"], "join")

    def test_not_joinable(self, group, authz_policy):
        group.joinable_by = None
        assert not authz_policy.permits(
            group, ["userid", "authority:example.com"], "join"
        )

    def test_world_readable(self, group, authz_policy):
        group.readable_by = ReadableBy.world
        assert authz_policy.permits(group, [security.Everyone], "read")

    def test_world_readable_grants_flag_permissions(self, group, authz_policy):
        group.readable_by = ReadableBy.world
        assert authz_policy.permits(group, [security.Authenticated], "flag")
        assert not authz_policy.permits(group, [security.Everyone], "flag")

    def test_world_readable_does_not_grant_moderate_permissions(
        self, group, authz_policy
    ):
        group.readable_by = ReadableBy.world
        assert not authz_policy.permits(group, [security.Authenticated], "moderate")
        assert not authz_policy.permits(group, [security.Everyone], "moderate")

    def test_members_readable(self, group, authz_policy):
        group.readable_by = ReadableBy.members
        assert authz_policy.permits(group, ["group:test-group"], "read")

    def test_members_flaggable(self, group, authz_policy):
        group.readable_by = ReadableBy.members
        assert authz_policy.permits(group, ["group:test-group"], "flag")

    def test_non_creator_members_do_not_have_moderate_permission(
        self, group, authz_policy
    ):
        group.readable_by = ReadableBy.members
        assert not authz_policy.permits(group, ["group:test-group"], "moderate")

    def test_not_readable(self, group, authz_policy):
        group.readable_by = None
        assert not authz_policy.permits(
            group, [security.Everyone, "group:test-group"], "read"
        )

    def test_not_flaggable(self, group, authz_policy):
        group.readable_by = None
        assert not authz_policy.permits(
            group, [security.Authenticated, "group:test-group"], "flag"
        )

    def test_authority_writeable(self, group, authz_policy):
        group.writeable_by = WriteableBy.authority
        assert authz_policy.permits(group, ["authority:example.com"], "write")

    def test_members_writeable(self, group, authz_policy):
        group.writeable_by = WriteableBy.members
        assert authz_policy.permits(group, ["group:test-group"], "write")

    def test_not_writeable(self, group, authz_policy):
        group.writeable_by = (None,)
        assert not authz_policy.permits(
            group, ["authority:example.com", "group:test-group"], "write"
        )

    def test_creator_has_moderate_permission(self, group, authz_policy):
        assert authz_policy.permits(group, "acct:luke@example.com", "moderate")

    def test_creator_has_admin_permission(self, group, authz_policy):
        assert authz_policy.permits(group, "acct:luke@example.com", "admin")

    def test_auth_client_with_matching_authority_has_admin_permission(
        self, group, authz_policy
    ):
        group.authority = "weewhack.com"

        assert authz_policy.permits(
            group, ["flip", "client_authority:weewhack.com"], "admin"
        )

    def test_auth_client_without_matching_authority_does_not_have_admin_permission(
        self, group, authz_policy
    ):
        group.authority = "weewhack.com"

        assert not authz_policy.permits(
            group, ["flip", "client_authority:2weewhack.com"], "admin"
        )

    def test_creator_has_upsert_permissions(self, group, authz_policy):
        assert authz_policy.permits(group, "acct:luke@example.com", "upsert")

    def test_admin_allowed_only_for_authority_when_no_creator(
        self, group, authz_policy
    ):
        group.creator = None

        principals = authz_policy.principals_allowed_by_permission(group, "admin")

        assert "client_authority:example.com" in principals
        assert authz_policy.permits(
            group, ["flip", "client_authority:example.com"], "admin"
        )

    def test_no_moderate_permission_when_no_creator(self, group, authz_policy):
        group.creator = None

        principals = authz_policy.principals_allowed_by_permission(group, "moderate")
        assert len(principals) == 0

    def test_no_upsert_permission_when_no_creator(self, group, authz_policy):
        group.creator = None

        principals = authz_policy.principals_allowed_by_permission(group, "upsert")
        assert len(principals) == 0

    def test_fallback_is_deny_all(self, group, authz_policy):
        assert not authz_policy.permits(group, [security.Everyone], "foobar")

    @pytest.fixture
    def authz_policy(self):
        return security.ACLAuthorizationPolicy()

    @pytest.fixture
    def group(self):
        creator = models.User(username="luke", authority="example.com")
        group = models.Group(
            name="test-group", authority="example.com", creator=creator
        )
        group.pubid = "test-group"
        return group



# -*- coding: utf-8 -*-

"""Custom SQLAlchemy types for use with the Annotations API database."""
from __future__ import unicode_literals

from hyputils.memex._compat import string_types
import binascii
import base64
import uuid

from sqlalchemy import types
from sqlalchemy.dialects import postgresql
from sqlalchemy.exc import DontWrapMixin


# A magic byte (expressed as two hexadecimal nibbles) which we use to expand a
# 15-byte ElasticSearch flake ID into a 16-byte UUID.
#
# The UUID specification defines UUIDs as taking the form
#
#     xxxxxxxx-xxxx-Mxxx-Nxxx-xxxxxxxxxxxx
#
# in the canonical hexadecimal representation. M and N represent the UUID
# version and variant fields respectively. The four bits M can take values {1,
# 2, 3, 4, 5} in specified UUID types, and the first three bits of N can take
# the values {8, 9, 0xa, 0xb} in specified UUID types.
#
# In order to expand a 15-byte ElasticSearch flake ID into a value that can be
# stored in the UUID field, we insert the magic nibbles 0xe, 0x5 into the
# version and variant fields respectively. These values are disjoint with any
# specified UUID so the resulting UUID can be distinguished from those
# generated by, for example, PostgreSQL's uuid_generate_v1mc(), and mapped back
# to a 20-char ElasticSearch flake ID.
ES_FLAKE_MAGIC_BYTE = ["e", "5"]


class InvalidUUID(Exception, DontWrapMixin):
    pass


class URLSafeUUID(types.TypeDecorator):

    """
    Expose UUIDs as URL-safe base64-encoded strings.

    Fields decorated with this type decorator use PostgreSQL UUID fields for
    storage, but expose URL-safe strings in the application.

    This type decorator will handle the transformation between any UUID and a
    URL-safe, base64-encoded string version of that UUID (which will be 22
    characters long). In addition, it will transparently map post-v1.4
    ElasticSearch flake IDs (which are 20 characters long and map to 15 bytes
    of data).
    """

    impl = postgresql.UUID

    def process_bind_param(self, value, dialect):
        if value is None:
            return None
        return _get_hex_from_urlsafe(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        hexstring = uuid.UUID(value).hex
        return _get_urlsafe_from_hex(hexstring)


class AnnotationSelectorJSONB(types.TypeDecorator):

    """
    Special type for the Annotation selector column.

    It transparently escapes NULL (\u0000) bytes to \\u0000 when writing to the
    database, and the other way around when reading from the database, but
    only on the prefix/exact/suffix fields in a TextQuoteSelector.
    """

    impl = postgresql.JSONB

    def process_bind_param(self, value, dialect):
        return _transform_quote_selector(value, _escape_null_byte)

    def process_result_value(self, value, dialect):
        return _transform_quote_selector(value, _unescape_null_byte)


def _get_hex_from_urlsafe(value):
    """
    Convert a URL-safe base 64 ID to a hex UUID.

    :type value: unicode
    :rtype: unicode
    """

    def _fail():
        raise InvalidUUID("{0!r} is not a valid encoded UUID".format(value))

    if not isinstance(value, string_types):
        raise InvalidUUID(
            "`value` is {}, expected one of {}".format(type(value), string_types)
        )

    bytestr = value.encode()

    if len(bytestr) == 22:
        # 22-char inputs represent 16 bytes of data, which when normally
        # base64-encoded would have two bytes of padding on the end, so we add
        # that back before decoding.
        try:
            data = _must_b64_decode(bytestr + b"==", expected_size=16)
        except (TypeError, binascii.Error):
            _fail()
        return binascii.hexlify(data).decode()

    if len(bytestr) == 20:
        # 20-char inputs represent 15 bytes of data, which requires no padding
        # corrections.
        try:
            data = _must_b64_decode(bytestr, expected_size=15)
        except (TypeError, binascii.Error):
            _fail()
        hexstring = binascii.hexlify(data).decode()
        # These are ElasticSearch flake IDs, so to convert them into UUIDs we
        # insert the magic nibbles at the appropriate points. See the comments
        # on ES_FLAKE_MAGIC_BYTE for details.
        return (
            hexstring[0:12]
            + ES_FLAKE_MAGIC_BYTE[0]
            + hexstring[12:15]
            + ES_FLAKE_MAGIC_BYTE[1]
            + hexstring[15:30]
        )

    # Fallthrough: we must have a received a string of invalid length
    _fail()


def _get_urlsafe_from_hex(value):
    """
    Convert a hex UUID to a URL-safe base 64 ID.

    :type value: unicode
    :rtype: unicode
    """

    # Validate and normalise hex string
    hexstring = uuid.UUID(hex=value).hex

    is_flake_id = (
        hexstring[12] == ES_FLAKE_MAGIC_BYTE[0]
        and hexstring[16] == ES_FLAKE_MAGIC_BYTE[1]
    )

    if is_flake_id:
        # The hex representation of the flake ID is simply the UUID without the
        # two magic nibbles.
        data = binascii.unhexlify(hexstring[0:12] + hexstring[13:16] + hexstring[17:32])
        return base64.urlsafe_b64encode(data).decode()

    # Encode UUID bytes and strip two bytes of padding
    data = binascii.unhexlify(hexstring)
    return base64.urlsafe_b64encode(data)[:-2].decode()


def _must_b64_decode(data, expected_size=None):
    result = base64.urlsafe_b64decode(data)
    if expected_size is not None and len(result) != expected_size:
        raise TypeError("incorrect data size")
    return result


def _transform_quote_selector(selectors, transform_func):
    if selectors is None:
        return None

    if not isinstance(selectors, list):
        return selectors

    for selector in selectors:
        if not isinstance(selector, dict):
            continue

        if not selector.get("type") == "TextQuoteSelector":
            continue

        if "prefix" in selector:
            selector["prefix"] = transform_func(selector["prefix"])
        if "exact" in selector:
            selector["exact"] = transform_func(selector["exact"])
        if "suffix" in selector:
            selector["suffix"] = transform_func(selector["suffix"])

    return selectors


def _escape_null_byte(s):
    if s is None:
        return s

    return s.replace("\u0000", "\\u0000")


def _unescape_null_byte(s):
    if s is None:
        return s

    return s.replace("\\u0000", "\u0000")



import re
from setuptools import setup


def find_version(filename):
    _version_re = re.compile(r"__version__ = '(.*)'")
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            return version_match.group(1)


__version__ = find_version('hyputils/__init__.py')

with open('README.md', 'rt') as f:
    long_description = f.read()

tests_memex_require = ['bleach',
                       'jsonschema',
                       'mistune',
                       'psycopg2',
                       'python-slugify',
                       'sqlalchemy',
                       'webob']
tests_require = ['factory-boy', 'mock', 'pytest', 'pytest-runner'] + tests_memex_require
setup(name='hyputils',
      version=__version__,
      description='Python utilities for the Hypothes.is REST api and websocket interface',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/tgbugs/hyputils',
      author='Tom Gillespie',
      author_email='tgbugs@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
      ],
      keywords='hypothesis hypothes.is web annotation',
      packages=['hyputils',
                'hyputils.memex',
                'hyputils.memex.db',
                'hyputils.memex.models',
                'hyputils.memex.schemas',
                'hyputils.memex.util',],
      python_requires='>=3.6',
      tests_require=tests_require,
      install_requires=[
          'appdirs',
          'certifi',
          'requests',
          'websockets',
      ],
      extras_require={'dev': ['pytest-cov', 'wheel'],
                      'memex':['python-dateutil'] + tests_memex_require,
                      'test': tests_require,
                      'zdesk': ['pyyaml', 'zdesk'],
                     },
      entry_points={
          'console_scripts': [
          ],
      },
     )



# -*- coding: utf-8 -*-

"""
Configure and expose the application database session.

This module is responsible for setting up the database session and engine, and
making that accessible to other parts of the application.

Models should inherit from `h.db.Base` in order to have their metadata bound at
application startup.

Most application code should access the database session using the request
property `request.db` which is provided by this module.
"""
from __future__ import unicode_literals

import logging

import sqlalchemy
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import exc
from sqlalchemy.orm import sessionmaker

emptyset = """<svg width="400" height="400" version="1.0" xmlns="http://www.w3.org/2000/svg">
<path d="m377.25 39.844-48.828 48.828c27.994 31.739 41.992 68.929 41.992 111.57-3.7e-4 47.038-16.683 87.199-50.049 120.48-33.366 33.285-73.568 49.927-120.61 49.927-42.481-1e-5 -79.671-13.835-111.57-41.504l-48.34 48.096-17.09-17.09 48.584-47.852c-27.995-32.226-41.992-69.58-41.992-112.06-2.9e-5 -47.038 16.642-87.199 49.927-120.48 33.284-33.284 73.445-49.926 120.48-49.927 42.643 3.36e-4 79.834 13.998 111.57 41.992l49.072-49.072 16.846 17.09zm-83.008 48.828c-27.018-23.274-58.513-34.912-94.482-34.912-40.202 3.12e-4 -74.666 14.364-103.39 43.091-28.727 28.727-43.091 63.192-43.091 103.39-5.3e-5 35.97 11.637 67.464 34.912 94.482l206.05-206.05zm52.002 111.57c-3.4e-4 -35.97-11.638-67.464-34.912-94.482l-206.05 206.05c27.018 23.275 58.512 34.912 94.482 34.912 40.202 1e-5 74.666-14.364 103.39-43.091 28.727-28.727 43.09-63.192 43.091-103.39z"/>
</svg>"""


__all__ = ("Base", "Session", "init", "make_engine")

log = logging.getLogger(__name__)

# Create a default metadata object with naming conventions for indexes and
# constraints. This makes changing such constraints and indexes with alembic
# after creation much easier. See:
#
#   http://docs.sqlalchemy.org/en/latest/core/constraints.html#configuring-constraint-naming-conventions
#
metadata = sqlalchemy.MetaData(
    naming_convention={
        "ix": "ix__%(column_0_label)s",
        "uq": "uq__%(table_name)s__%(column_0_name)s",
        "ck": "ck__%(table_name)s__%(constraint_name)s",
        "fk": "fk__%(table_name)s__%(column_0_name)s__%(referred_table_name)s",
        "pk": "pk__%(table_name)s",
    }
)

Base = declarative_base(metadata=metadata)

Session = sessionmaker()


def init(engine, base=Base, should_create=False, should_drop=False, authority=None):
    """Initialise the database tables managed by `h.db`."""
    # Import models package to populate the metadata
    import hyputils.memex.models  # noqa

    if should_drop:
        base.metadata.reflect(engine)
        base.metadata.drop_all(engine)
    if should_create:
        # In order to be able to generate UUIDs, we load the uuid-ossp
        # extension.
        engine.execute('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";')
        base.metadata.create_all(engine)

    default_org = _maybe_create_default_organization(engine, authority)
    _maybe_create_world_group(engine, authority, default_org)


def make_engine(settings):
    """Construct a sqlalchemy engine from the passed ``settings``."""
    return sqlalchemy.create_engine(settings["sqlalchemy.url"])


def _maybe_create_default_organization(engine, authority, logopath=None):
    from hyputils.memex import models
    if logopath is None:
        logo = emptyset
    else:
        with open(logopath, 'rb') as f:
            logo = f.read().decode("utf-8")

    session = Session(bind=engine)

    try:
        default_org = models.Organization.default(session)
    except exc.NoResultFound:
        default_org = None

    if default_org is None:
        default_org = models.Organization(
            name="Hypothesis", authority=authority, pubid="__default__"
        )
        default_org.logo = logo
        session.add(default_org)

    session.commit()
    session.close()

    return default_org


def _maybe_create_world_group(engine, authority, default_org):
    from hyputils.memex import models
    from hyputils.memex.models.group import ReadableBy, WriteableBy

    session = Session(bind=engine)
    world_group = session.query(models.Group).filter_by(pubid="__world__").one_or_none()
    if world_group is None:
        world_group = models.Group(
            name="Public",
            authority=authority,
            joinable_by=None,
            readable_by=ReadableBy.world,
            writeable_by=WriteableBy.authority,
            organization=default_org,
        )
        world_group.pubid = "__world__"
        session.add(world_group)

    session.commit()
    session.close()



import unittest
from pathlib import Path
from hyputils.hypothesis import Memoizer
from hyputils.hypothesis import api_token, username, group, group_to_memfile

bad_memfile = 'test-bad-memfile.json'


class TestMemoize(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.get_annos = Memoizer(group_to_memfile(group), api_token, username, group)

        world_get = Memoizer(bad_memfile, api_token, username, '__world__')
        world_annos = world_get.get_annos_from_api(max_results=10)
        world_get.memoize_annos(world_annos)

    @classmethod
    def tearDownClass(cls):
        Path(bad_memfile).unlink()

    def test_0_get_start(self):
        annos = self.get_annos.get_annos_from_api(max_results=400)
        self.get_annos.memoize_annos(annos)

    def test_1_get_file(self):
        annos, lsu = self.get_annos.get_annos_from_file()

    def test_2_add_missin(self):
        annos, lsu = self.get_annos.get_annos_from_file()
        more_annos = self.get_annos.add_missing_annos(annos, lsu)
        assert len(more_annos) > 800 > len(annos)

    def test_3_get_rest(self):
        annos = self.get_annos.get_annos_from_api(max_results=400)
        self.get_annos.memoize_annos(annos)
        more_annos = self.get_annos()
        assert len(more_annos) > 800 > len(annos)

    def test_4_group_mismatch_at_load_from_file(self):
        group_get = Memoizer(bad_memfile, api_token, username, group)
        try:
            world_annos = group_get.get_annos_from_file()
            raise AssertionError('should have failed due to group mismatch with __world__')
        except Memoizer.GroupMismatchError as e:
            pass

    def test_5_group_mismatch_at_add_missing(self):
        group_get = Memoizer(bad_memfile, api_token, username, group)
        try:
            group_get.get_annos()
            raise AssertionError('should have failed due to group mismatch with __world__')
        except Memoizer.GroupMismatchError as e:
            pass



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from hyputils.memex._compat import PY2
import enum
from mock import Mock
import pytest

from hyputils.memex.schemas import ValidationError
from hyputils.memex.schemas.base import JSONSchema


class ExampleJSONSchema(JSONSchema):
    # Use `bytes` for property names in Py 2 so that exception messages about
    # missing properties have the same content in Py 2 + Py 3.
    prop_name_type = bytes if PY2 else str

    schema = {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "properties": {
            prop_name_type("foo"): {"type": "string"},
            prop_name_type("bar"): {"type": "integer"},
        },
        "required": [prop_name_type("foo"), prop_name_type("bar")],
    }


class TestJSONSchema(object):
    def test_it_returns_data_when_valid(self):
        data = {"foo": "baz", "bar": 123}

        assert ExampleJSONSchema().validate(data) == data

    def test_it_raises_when_data_invalid(self):
        data = 123  # not an object

        with pytest.raises(ValidationError):
            ExampleJSONSchema().validate(data)

    def test_it_sets_appropriate_error_message_when_data_invalid(self):
        data = {"foo": "baz"}  # required bar is missing

        with pytest.raises(ValidationError) as e:
            ExampleJSONSchema().validate(data)

        message = str(e.value)
        assert message.startswith("'bar' is a required property")

    def test_it_returns_all_errors_in_message(self):
        data = {}  # missing both required fields

        with pytest.raises(ValidationError) as e:
            ExampleJSONSchema().validate(data)

        message = str(e.value)
        assert message.startswith(
            "'foo' is a required property, 'bar' is a required property"
        )
